# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Plugins manager dialog"""

import os.path
import logging
from ui.qt import (Qt, pyqtSignal, QDialog, QTreeWidgetItem, QTreeWidget,
                   QVBoxLayout, QTextEdit, QDialogButtonBox, QLabel,
                   QFontMetrics, QHeaderView, QPushButton)
from ui.itemdelegates import NoOutlineHeightDelegate
from utils.pixmapcache import getIcon


STATE_COL = 0       # Enabled/disabled
CONFLICT_COL = 1    # Exclamation sign
TYPE_COL = 2        # Sys/user
NAME_COL = 3
VERSION_COL = 4
SETTINGS_COL = 5    # Settings button


class SettingsButton(QPushButton):

    """Custom settings button"""

    CustomClick = pyqtSignal(int)

    def __init__(self):
        QPushButton.__init__(self, getIcon('pluginsettings.png'), "")
        self.setFixedSize(24, 24)
        self.setFocusPolicy(Qt.NoFocus)

        self.index = -1
        self.clicked.connect(self.onClick)

    def onClick(self):
        """Emits a signal with the button index"""
        self.CustomClick.emit(self.index)


class PluginItem(QTreeWidgetItem):

    """Single plugin item"""

    def __init__(self, pluginManager, cdmPlugin, active, category):
        self.plugin = cdmPlugin
        self.active = active
        self.category = category

        name = self.plugin.getName()
        ver = self.plugin.getVersion()
        QTreeWidgetItem.__init__(self, ["", "", "", name, ver])

        if not self.plugin.conflictType in [pluginManager.NO_CONFLICT,
                                            pluginManager.USER_DISABLED]:
            self.setIcon(CONFLICT_COL, getIcon('pluginconflict.png'))
            self.setToolTip(CONFLICT_COL, self.plugin.conflictMessage)

        self.setToolTip(STATE_COL, "Enable / disable")
        self.setToolTip(CONFLICT_COL, "Conflict")

        if self.plugin.isUser():
            self.setIcon(TYPE_COL, getIcon('pluginuser.png'))
            self.setToolTip(TYPE_COL, "User plugin")
        else:
            self.setIcon(TYPE_COL, getIcon('pluginsystem.png'))
            self.setToolTip(TYPE_COL, "System wide plugin")

        self.setFlags(self.flags() | Qt.ItemIsUserCheckable)
        if active:
            self.setCheckState(STATE_COL, Qt.Checked)
        else:
            self.setCheckState(STATE_COL, Qt.Unchecked)


class PluginsDialog(QDialog):

    """Codimension plugins dialog"""

    def __init__(self, pluginManager, parent=None):
        QDialog.__init__(self, parent)
        self.setWindowTitle("Plugin Manager")

        self.__pluginManager = pluginManager
        self.__configFuncs = {} # int -> callable

        self.__createLayout()
        self.__populate()

        self.__pluginsView.setFocus()
        self.__inItemChange = False

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(640, 480)
        self.setSizeGripEnabled(True)

        layout = QVBoxLayout()

        # Plugins list
        self.__pluginsView = QTreeWidget()
        self.__pluginsView.setAlternatingRowColors(True)
        self.__pluginsView.setRootIsDecorated(False)
        self.__pluginsView.setItemsExpandable(False)
        self.__pluginsView.setSortingEnabled(True)
        self.__pluginsView.setItemDelegate(NoOutlineHeightDelegate(4))
        self.__pluginsView.setUniformRowHeights(True)

        # Alert | system/user | Enable | Name | Version
        self.__pluginsHeader = QTreeWidgetItem(["", "", "",
                                                "Name", "Version", ""])
        self.__pluginsView.setHeaderItem(self.__pluginsHeader)
        self.__pluginsView.header().setSortIndicator(NAME_COL,
                                                     Qt.AscendingOrder)
        self.__pluginsView.itemSelectionChanged.connect(
            self.__pluginSelectionChanged)
        self.__pluginsView.itemChanged.connect(self.__onItemChanged)

        layout.addWidget(self.__pluginsView)

        # Detailed information
        detailsLabel = QLabel("Detailed information")
        layout.addWidget(detailsLabel)
        self.__details = QTreeWidget()
        self.__details.setAlternatingRowColors(False)
        self.__details.setRootIsDecorated(False)
        self.__details.setItemsExpandable(False)
        self.__details.setSortingEnabled(False)
        self.__details.setItemDelegate(NoOutlineHeightDelegate(4))
        self.__details.setUniformRowHeights(True)

        detailsHeader = QTreeWidgetItem(["", ""])
        self.__details.setHeaderItem(detailsHeader)
        self.__details.setHeaderHidden(True)

        metrics = QFontMetrics(self.__details.font())
        rect = metrics.boundingRect("X")
        self.__details.setFixedHeight(rect.height() * 6 + 5)
        layout.addWidget(self.__details)

        # Errors/warnings
        errorsLabel = QLabel("Errors / warnings")
        layout.addWidget(errorsLabel)
        self.__errorsText = QTextEdit()
        self.__errorsText.setReadOnly(True)
        self.__errorsText.setAcceptRichText(False)
        metrics = QFontMetrics(self.__errorsText.font())
        rect = metrics.boundingRect("X")
        self.__errorsText.setFixedHeight(rect.height() * 4 + 5)
        layout.addWidget(self.__errorsText)

        # Buttons box
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Ok)
        self.__OKButton = buttonBox.button(QDialogButtonBox.Ok)
        self.__OKButton.setDefault(True)
        buttonBox.accepted.connect(self.close)
        buttonBox.rejected.connect(self.close)
        layout.addWidget(buttonBox)

        self.setLayout(layout)

    def __createConfigButton(self):
        """Creates a configure button for a plugin"""
        button = SettingsButton()
        button.CustomClick.connect(self.onPluginSettings)
        return button

    def __populate(self):
        """Populates the list with the plugins"""
        index = 0

        for category in self.__pluginManager.activePlugins:
            for cdmPlugin in self.__pluginManager.activePlugins[category]:
                newItem = PluginItem(self.__pluginManager, cdmPlugin,
                                     True, category)
                self.__pluginsView.addTopLevelItem(newItem)
                settingsButton = self.__createConfigButton()

                try:
                    configFunction = cdmPlugin.getObject().getConfigFunction()
                    if configFunction is None:
                        settingsButton.setToolTip(
                            "Plugin does not need configuring")
                        settingsButton.setEnabled(False)
                    else:
                        settingsButton.setToolTip("Click to configure")
                        settingsButton.setEnabled(True)
                        self.__configFuncs[index] = configFunction
                        settingsButton.index = index
                        index += 1
                except Exception:
                    settingsButton.setToolTip("Bad plugin interface. No "
                                              "configuration function "
                                              "received.")
                    settingsButton.setEnabled(False)

                self.__pluginsView.setItemWidget(newItem, SETTINGS_COL,
                                                 settingsButton)

        for category in self.__pluginManager.inactivePlugins:
            for cdmPlugin in self.__pluginManager.inactivePlugins[category]:
                newItem = PluginItem(self.__pluginManager, cdmPlugin,
                                     False, category)
                self.__pluginsView.addTopLevelItem(newItem)
                settingsButton = self.__createConfigButton()

                try:
                    configFunction = cdmPlugin.getObject().getConfigFunction()
                    if configFunction is None:
                        settingsButton.setToolTip(
                            "Plugin does not need configuring")
                        settingsButton.setEnabled(False)
                    else:
                        settingsButton.setToolTip(
                            "Enable plugin and then click to configure")
                        settingsButton.setEnabled(False)
                        self.__configFuncs[index] = configFunction
                        settingsButton.index = index
                        index += 1
                except:
                    settingsButton.setToolTip("Bad plugin interface. No "
                                              "configuration function "
                                              "received.")
                    settingsButton.setEnabled(False)

                self.__pluginsView.setItemWidget(newItem, SETTINGS_COL,
                                                 settingsButton)

        for cdmPlugin in self.__pluginManager.unknownPlugins:
            newItem = PluginItem(self.__pluginManager, cdmPlugin, False, None)
            self.__pluginsView.addTopLevelItem(newItem)
            settingsButton = self.__createConfigButton()
            settingsButton.setToolTip("Unknown plugins are not configurable")
            settingsButton.setEnabled(False)
            self.__pluginsView.setItemWidget(newItem, SETTINGS_COL,
                                             settingsButton)

        self.__sortPlugins()
        self.__resizePlugins()

    def __sortPlugins(self):
        """Sorts the plugins table"""
        self.__pluginsView.sortItems(
            self.__pluginsView.sortColumn(),
            self.__pluginsView.header().sortIndicatorOrder())

    def __resizePlugins(self):
        """Resizes the plugins table"""
        self.__pluginsView.header().setStretchLastSection(False)
        self.__pluginsView.header().resizeSections(
            QHeaderView.ResizeToContents)
        self.__pluginsView.header().resizeSection(STATE_COL, 28)
        self.__pluginsView.header().setSectionResizeMode(STATE_COL,
                                                         QHeaderView.Fixed)
        self.__pluginsView.header().resizeSection(CONFLICT_COL, 28)
        self.__pluginsView.header().setSectionResizeMode(CONFLICT_COL,
                                                         QHeaderView.Fixed)
        self.__pluginsView.header().resizeSection(TYPE_COL, 28)
        self.__pluginsView.header().setSectionResizeMode(TYPE_COL,
                                                         QHeaderView.Fixed)

        self.__pluginsView.header().setSectionResizeMode(VERSION_COL,
                                                         QHeaderView.Stretch)
        self.__pluginsView.header().resizeSection(SETTINGS_COL, 24)
        self.__pluginsView.header().setSectionResizeMode(SETTINGS_COL,
                                                         QHeaderView.Fixed)

    def __pluginSelectionChanged(self):
        """Triggered when an item is selected"""
        selected = list(self.__pluginsView.selectedItems())
        if selected:
            self.__updateDetails(selected[0])
        else:
            self.__updateDetails(None)

    def __updateDetails(self, item):
        """Updates the content of the details and the error boxes"""
        self.__details.clear()
        self.__errorsText.setText("")

        if item is None:
            return

        self.__details.addTopLevelItem(
            QTreeWidgetItem(["Author", item.plugin.getAuthor()]))
        self.__details.addTopLevelItem(
            QTreeWidgetItem(["Path", os.path.normpath(item.plugin.getPath())]))
        self.__details.addTopLevelItem(
            QTreeWidgetItem(["Description", item.plugin.getDescription()]))
        self.__details.addTopLevelItem(
            QTreeWidgetItem(["Web site", item.plugin.getWebsite()]))

        copyright = item.plugin.getCopyright()
        if copyright is not None:
            if copyright.lower() != "unknown":
                self.__details.addTopLevelItem(
                    QTreeWidgetItem(["Copyright", copyright]))

        for name in item.plugin.getDetails():
            value = item.plugin.getDetails()[name]
            self.__details.addTopLevelItem(QTreeWidgetItem([name, value]))

        self.__errorsText.setText(item.plugin.conflictMessage)

    def __onItemChanged(self, item, column):
        """Triggered when an item is changed"""
        if self.__inItemChange:
            return

        if item.active:
            self.__inItemChange = True
            item.plugin.disable()
            item.active = False

            settingsButton = self.__pluginsView.itemWidget(item, SETTINGS_COL)
            settingsButton.setEnabled(False)
            if settingsButton.index != -1:
                settingsButton.setToolTip(
                    "Enable plugin and then click to configure")

            if item.category in self.__pluginManager.inactivePlugins:
                self.__pluginManager. \
                    inactivePlugins[item.category].append(item.plugin)
            else:
                self.__pluginManager. \
                    inactivePlugins[item.category] = [item.plugin]
            self.__pluginManager. \
                activePlugins[item.category].remove(item.plugin)
            self.__pluginManager.saveDisabledPlugins()
            self.__inItemChange = False
            self.__pluginManager.sendPluginDeactivated(item.plugin)
            return

        self.__inItemChange = True
        message = self.__pluginManager.checkConflict(item.plugin)
        if message is not None:
            item.setCheckState(STATE_COL, Qt.Unchecked)
            self.__errorsText.setText(message)
            self.__inItemChange = False
            return

        try:
            item.plugin.enable()
            item.active = True
            if item.category in self.__pluginManager.activePlugins:
                self.__pluginManager. \
                    activePlugins[item.category].append(item.plugin)
            else:
                self.__pluginManager. \
                    activePlugins[item.category] = [item.plugin]
            self.__pluginManager. \
                inactivePlugins[item.category].remove(item.plugin)
            self.__pluginManager.saveDisabledPlugins()
            self.__errorsText.setText("")
            item.setIcon(CONFLICT_COL, getIcon('empty.png'))
            item.setToolTip(CONFLICT_COL, "")

            settingsButton = self.__pluginsView.itemWidget(item, SETTINGS_COL)
            if settingsButton.index != -1:
                settingsButton.setToolTip("Click to configure")
                settingsButton.setEnabled(True)
            self.__pluginManager.sendPluginActivated(item.plugin)
        except Exception as exc:
            item.setCheckState(STATE_COL, Qt.Unchecked)
            self.__errorsText.setText(
                "Error activating the plugin, exception is generated:\n" +
                str(exc))
        except:
            item.setCheckState(STATE_COL, Qt.Unchecked)
            self.__errorsText.setText(
                "Error activating the plugin, unknown exception")

        self.__inItemChange = False

    def onPluginSettings(self, index):
        """Triggered when a configuring function is called"""
        if index not in self.__configFuncs:
            return

        try:
            self.__configFuncs[index]()
        except Exception as exc:
            logging.error("Error calling the plugin configuration function. "
                          "Message: " + str(exc))



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Global data singleton"""

# pylint: disable=W0702
# pylint: disable=W0703

import os
from os.path import sep, realpath, isdir, exists, isfile
from distutils.spawn import find_executable
from plugins.manager.pluginmanager import CDMPluginManager
from .project import CodimensionProject
from .briefmodinfocache import BriefModuleInfoCache
from .settings import SETTINGS_DIR


# This function needs to have a rope project built smart
def getSubdirs(path, baseNamesOnly=True, excludePythonModulesDirs=True):
    """Provides a list of sub directories for the given path"""
    subdirs = []
    try:
        path = realpath(path) + sep
        for item in os.listdir(path):
            candidate = path + item
            if isdir(candidate):
                if excludePythonModulesDirs:
                    modFile = candidate + sep + "__init__.py"
                    if exists(modFile):
                        continue
                if baseNamesOnly:
                    subdirs.append(item)
                else:
                    subdirs.append(candidate)
    except:
        pass
    return subdirs


class GlobalDataWrapper:

    """Global data singleton"""

    def __init__(self):
        self.application = None
        self.mainWindow = None
        self.skin = None
        self.screenWidth = 0
        self.screenHeight = 0
        self.version = "unknown"
        self.originalSysPath = None
        self.project = CodimensionProject()

        self.pluginManager = CDMPluginManager()

        self.briefModinfoCache = BriefModuleInfoCache()

        self.graphvizAvailable = self.__checkGraphviz()
        self.javaAvailable = self.__checkJava()
        self.hexdumpAvailable = self.__checkHexdump()

    def getProfileOutputPath(self, procuuid):
        """Provides the path to the profile output file"""
        if self.project.isLoaded():
            return self.project.userProjectDir + procuuid + '.profile.out'

        # No project loaded
        return SETTINGS_DIR + procuuid + '.profile.out'

    def getProjectImportDirs(self):
        """Provides a list of the project import dirs if so"""
        if not self.project.isLoaded():
            return []
        return self.project.getImportDirsAsAbsolutePaths()

    def isProjectScriptValid(self):
        """True if the project script valid"""
        scriptName = self.project.getProjectScript()
        if not scriptName:
            return False
        if not exists(scriptName):
            return False
        if not isfile(scriptName):
            return False

        from .fileutils import isPythonFile
        return isPythonFile(scriptName)

    def getFileLineDocstring(self, fName, line):
        """Provides a docstring if so for the given file and line"""
        from .fileutils import isPythonFile
        if not isPythonFile(fName):
            return ''

        infoCache = self.briefModinfoCache

        def checkFuncObject(obj, line):
            """Checks docstring for a function or a class"""
            if obj.line == line or obj.keywordLine == line:
                if obj.docstring is None:
                    return True, ''
                return True, obj.docstring.text
            for item in obj.classes + obj.functions:
                found, docstring = checkFuncObject(item, line)
                if found:
                    return True, docstring
            return False, ''

        try:
            info = infoCache.get(fName)
            for item in info.classes + info.functions:
                found, docstring = checkFuncObject(item, line)
                if found:
                    return docstring
        except:
            pass
        return ''

    def getModInfo(self, path):
        """Provides a module info for the given file"""
        from .fileutils import isPythonFile
        if not isPythonFile(path):
            raise Exception('Trying to parse non-python file: ' + path)
        return self.briefModinfoCache.get(path)

    @staticmethod
    def __checkGraphviz():
        """Checks if graphviz is available"""
        return find_executable('dot') is not None

    @staticmethod
    def __checkJava():
        """Checks if java is available"""
        return find_executable('java') is not None

    @staticmethod
    def __checkHexdump():
        """Checks if hexdump is available"""
        return find_executable('hexdump') is not None


globalsSingleton = GlobalDataWrapper()


def GlobalData():
    """Global singleton access"""
    return globalsSingleton



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""codimension pixmap cache"""

# pylint: disable=W0702
# pylint: disable=W0703
# pylint: disable=C0305

from os.path import dirname, realpath, sep, isabs, exists
import sys
from ui.qt import QPixmap, QIcon


class PixmapCache():

    """pixmap cache"""

    def __init__(self):
        self.__cache = {}
        self.__locationCache = {}
        self.__searchDirs = None

    def __initSearchDirs(self):
        """Initializes the search directories list"""
        # There are a few cases here:
        # - default skin does not have a directory in the installation package
        # - Codimension additional skins have a directory in the installation
        #   package
        # - User custom skins do not have directories in the installation
        #   package
        # All type of skins may have a directory in ~/.codimension3/skins

        self.__searchDirs = []

        from utils.globals import GlobalData
        skin = GlobalData().skin

        # First priority search dir is the ~/.codimension3/skins/<name>
        skinDir = skin.getUserDir()
        if skinDir:
            if exists(skinDir):
                self.__searchDirs.append(skinDir)

        # Second piority is the dir where the skin came from
        skinDir = skin.getDir()
        if skinDir:
            if exists(skinDir):
                if skinDir not in self.__searchDirs:
                    self.__searchDirs.append(skinDir)

        # Third priority is the default location of the pixmaps (installation
        # package)
        self.__searchDirs.append(dirname(realpath(sys.argv[0])) + sep +
                                 'pixmaps' + sep)

    def __getPath(self, path):
        """Provides an absolute path"""
        if isabs(path):
            return path if exists(path) else None

        if self.__searchDirs is None:
            self.__initSearchDirs()

        for dirName in self.__searchDirs:
            fullPath = dirName + path
            if exists(fullPath):
                return fullPath
        return None

    def getPixmap(self, name):
        """Provides the required pixmap"""
        try:
            return self.__cache[name]
        except KeyError:
            path = self.__getPath(name)
            if path is None:
                pixmap = QPixmap()
                self.__cache[name] = pixmap
                return pixmap

            try:
                pixmap = QPixmap(path)
            except:
                pixmap = QPixmap()
            self.__cache[name] = pixmap
            return pixmap

    def getIcon(self, name):
        """Provides a pixmap as an icon"""
        return QIcon(self.getPixmap(name))

    def getLocation(self, name):
        """Provides the pixmap location (svg items require a path)"""
        try:
            return self.__locationCache[name]
        except KeyError:
            # The path could be None if not found anywhere
            path = self.__getPath(name)
            self.__locationCache[name] = path
            return path

# Pixmap cache: should be only one
# Access functions are below
PIXMAP_CACHE = PixmapCache()


def getIcon(name):
    """Syntactic shugar"""
    return PIXMAP_CACHE.getIcon(name)

def getPixmap(name):
    """Syntactic shugar"""
    return PIXMAP_CACHE.getPixmap(name)

def getPixmapLocation(name):
    """Syntactic shugar"""
    return PIXMAP_CACHE.getLocation(name)




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Redirected IO margin"""

from html import escape
from ui.qt import QWidget, QPainter, Qt, QToolTip
from qutepart.margins import MarginBase
from utils.misc import extendInstance
from utils.globals import GlobalData
from utils.colorfont import getZoomedMarginFont
from .redirectedmsg import IOConsoleMsg


class CDMRedirectedIOMargin(QWidget):

    """Redirected IO margin widget"""

    _LEFT_MARGIN = 5
    _RIGHT_MARGIN = 3

    # Prefixes should match the message rendering (when copied to clipboard)
    MSG_TYPE_PROPS = {IOConsoleMsg.IDE_MESSAGE: ['ide ', None],
                      IOConsoleMsg.STDOUT_MESSAGE: ['out ', None],
                      IOConsoleMsg.STDERR_MESSAGE: ['err ', None],
                      IOConsoleMsg.STDIN_MESSAGE: [' in ', None]}

    def __init__(self, parent):
        QWidget.__init__(self, parent)

        # Margin data is:
        # {lineNo: [(text, tooltip, msgType),...], }
        # line number is 1 based, tooltip and fgColor could be None
        self.__data = {}

        extendInstance(self, MarginBase)
        MarginBase.__init__(self, parent, 'cdm_redirected_io_margin', 0)
        self.setMouseTracking(True)

        skin = GlobalData().skin
        self.__bgColor = skin['marginPaper']
        self.__fgColor = skin['marginColor']   # default

        CDMRedirectedIOMargin.MSG_TYPE_PROPS[IOConsoleMsg.IDE_MESSAGE][1] = \
            skin['ioconsoleMarginIDEMsgColor']
        CDMRedirectedIOMargin.MSG_TYPE_PROPS[IOConsoleMsg.STDOUT_MESSAGE][1] = \
            skin['ioconsoleMarginStdoutColor']
        CDMRedirectedIOMargin.MSG_TYPE_PROPS[IOConsoleMsg.STDERR_MESSAGE][1] = \
            skin['ioconsoleMarginStderrColor']
        CDMRedirectedIOMargin.MSG_TYPE_PROPS[IOConsoleMsg.STDIN_MESSAGE][1] = \
            skin['ioconsoleMarginStdinColor']

        self.__width = self.__calculateWidth()
        self.onTextZoomChanged()

        # The width needs to be re-calculated when the margin is drawn the
        # first time. The problem is that if the widget is not on the screen
        # then the font metrics are not calculated properly and thus the width
        # is not shown right. What I observed is an offset up to 2 pixels.
        self.__firstTime = True

    def __updateWidth(self):
        """Updates the margin width"""
        newWidth = self.__calculateWidth()
        if newWidth != self.__width:
            self.__width = newWidth
            self._qpart.updateViewport()

    def paintEvent(self, event):
        """Paints the margin"""
        if self.__firstTime:
            self.__updateWidth()
            self.__firstTime = False

        painter = QPainter(self)
        painter.fillRect(event.rect(), self.__bgColor)

        block = self._qpart.firstVisibleBlock()
        blockNumber = block.blockNumber()
        top = int(self._qpart.blockBoundingGeometry(block).
                  translated(self._qpart.contentOffset()).top())
        bottom = top + int(self._qpart.blockBoundingRect(block).height())

        boundingRect = self._qpart.blockBoundingRect(block)
        availableWidth = self.__width - self._RIGHT_MARGIN - self._LEFT_MARGIN

        # The margin font could be smaller than the main area font
        topShift = int((self._qpart.fontMetrics().height() -
                        self.fontMetrics().height()) / 2)
        if topShift < 0:
            topShift = 0

        availableHeight = self._qpart.fontMetrics().height()
        while block.isValid() and top <= event.rect().bottom():
            if block.isVisible() and bottom >= event.rect().top():
                lineno = blockNumber + 1
                props = self.__data.get(lineno, [(None, None, None)])
                text = props[0][0]
                if text:
                    msgType = props[0][2]
                    color = CDMRedirectedIOMargin.MSG_TYPE_PROPS[msgType][1]
                    if color is None:
                        color = self.__fgColor
                    painter.setPen(color)
                    painter.drawText(self._LEFT_MARGIN, top + topShift,
                                     availableWidth,
                                     availableHeight,
                                     Qt.AlignRight, text)
            block = block.next()
            boundingRect = self._qpart.blockBoundingRect(block)
            top = bottom
            bottom = top + int(boundingRect.height())
            blockNumber += 1

    def mouseMoveEvent(self, event):
        """Tooltips for the text"""
        blockNumber = self._qpart.cursorForPosition(
            event.pos()).blockNumber()
        lineno = blockNumber + 1
        if lineno in self.__data:
            props = self.__data.get(lineno, [(None, None, None)])
            tooltipItems = []
            for prop in props:
                if prop[1]:
                    msgType = prop[2]
                    tooltipItems.append(
                        CDMRedirectedIOMargin.MSG_TYPE_PROPS[msgType][0] +
                        prop[1])
            if tooltipItems:
                tooltip = "<p style='white-space:pre'>" + \
                          escape('\n'.join(tooltipItems)) + "</p>"
            QToolTip.showText(event.globalPos(), tooltip)
        else:
            QToolTip.hideText()
        return QWidget.mouseMoveEvent(self, event)

    def __calculateWidth(self):
        """Calculates the margin width"""
        digits = len('88:88:88.888')   # Basically it is a timestamp
        digitsWidth = self.fontMetrics().width('9') * digits
        return self._LEFT_MARGIN + digitsWidth + self._RIGHT_MARGIN

    def width(self):
        """Desired width. Includes text and margins"""
        return self.__width

    def setFont(self, font):
        """Overloaded to adjust the width if needed"""
        QWidget.setFont(self, font)
        self.__updateWidth()

    def setBackgroundColor(self, color):
        """Sets the new background color"""
        if self.__bgColor != color:
            self.__bgColor = color
            self.update()

    def onTextZoomChanged(self):
        """Triggered when a zoom has been changed"""
        self.setFont(getZoomedMarginFont())

    def clear(self):
        """Clears all the data"""
        if self.__data:
            self.__data = {}
            self.update()

    def addData(self, lineno, text, tooltip, msgType):
        """Appends new data; lineno is 1-based"""
        if lineno in self.__data:
            self.__data[lineno].append((text, tooltip, msgType))
        else:
            self.__data[lineno] = [(text, tooltip, msgType)]
            self.update()

    def getLineMessageType(self, lineno):
        """Provides the last line message type"""
        if lineno in self.__data:
            return self.__data[lineno][0][2]
        return None



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Sidebar implementation"""

from .qt import (QEvent, QSize, Qt, QTabBar, QWidget, QStackedWidget,
                 QBoxLayout, pyqtSignal)


class SideBar(QWidget):

    """Sidebar with a widget area which is hidden or shown.

    On by clicking any tab, off by clicking the current tab.
    """

    North = 0
    East = 1
    South = 2
    West = 3

    sigTabCloseRequested = pyqtSignal(int)

    def __init__(self, orientation, parent=None):
        QWidget.__init__(self, parent)

        self.__tabBar = QTabBar()
        self.__tabBar.setDrawBase(True)
        self.__tabBar.setFocusPolicy(Qt.NoFocus)
        self.__tabBar.setUsesScrollButtons(True)
        self.__tabBar.setElideMode(1)
        self.__tabBar.tabCloseRequested.connect(self.__onCloseRequest)
        self.__stackedWidget = QStackedWidget(self)
        self.__stackedWidget.setContentsMargins(0, 0, 0, 0)
        self.barLayout = QBoxLayout(QBoxLayout.LeftToRight)
        self.barLayout.setContentsMargins(0, 0, 0, 0)
        self.layout = QBoxLayout(QBoxLayout.TopToBottom)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.barLayout.addWidget(self.__tabBar)
        self.layout.addLayout(self.barLayout)
        self.layout.addWidget(self.__stackedWidget)
        self.setLayout(self.layout)

        self.__minimized = False
        self.__minSize = 0
        self.__maxSize = 0
        self.__bigSize = QSize()

        self.splitter = None

        self.__tabBar.installEventFilter(self)

        self.__orientation = orientation
        self.setOrientation(orientation)

        self.__tabBar.currentChanged.connect(
            self.__stackedWidget.setCurrentIndex)

    def setSplitter(self, splitter):
        """Set the splitter managing the sidebar"""
        self.splitter = splitter

    def __getIndex(self):
        """Provides the widget index in splitters"""
        if self.__orientation == SideBar.West:
            return 0
        if self.__orientation == SideBar.East:
            return 2
        if self.__orientation == SideBar.South:
            return 1
        return 0

    def shrink(self):
        """Shrink the sidebar"""
        if self.__minimized:
            return

        self.__minimized = True
        self.__bigSize = self.size()
        if self.__orientation in [SideBar.North, SideBar.South]:
            self.__minSize = self.minimumHeight()
            self.__maxSize = self.maximumHeight()
        else:
            self.__minSize = self.minimumWidth()
            self.__maxSize = self.maximumWidth()

        self.__stackedWidget.hide()

        sizes = self.splitter.sizes()
        selfIndex = self.__getIndex()

        if self.__orientation in [SideBar.North, SideBar.South]:
            newHeight = self.__tabBar.minimumSizeHint().height()
            self.setFixedHeight(newHeight)

            diff = sizes[selfIndex] - newHeight
            sizes[selfIndex] = newHeight
        else:
            newWidth = self.__tabBar.minimumSizeHint().width()
            self.setFixedWidth(newWidth)

            diff = sizes[selfIndex] - newWidth
            sizes[selfIndex] = newWidth

        if selfIndex == 0:
            sizes[1] += diff
        else:
            sizes[selfIndex - 1] += diff

        self.splitter.setSizes(sizes)

    def expand(self):
        """Expand the sidebar"""
        if not self.__minimized:
            return

        self.__minimized = False
        self.__stackedWidget.show()
        self.resize(self.__bigSize)

        sizes = self.splitter.sizes()
        selfIndex = self.__getIndex()

        if self.__orientation in [SideBar.North, SideBar.South]:
            self.setMinimumHeight(self.__minSize)
            self.setMaximumHeight(self.__maxSize)

            diff = self.__bigSize.height() - sizes[selfIndex]
            sizes[selfIndex] = self.__bigSize.height()
        else:
            self.setMinimumWidth(self.__minSize)
            self.setMaximumWidth(self.__maxSize)

            diff = self.__bigSize.width() - sizes[selfIndex]
            sizes[selfIndex] = self.__bigSize.width()

        if selfIndex == 0:
            sizes[1] -= diff
        else:
            sizes[selfIndex - 1] -= diff

        self.splitter.setSizes(sizes)

    def isMinimized(self):
        """Provides the minimized state"""
        return self.__minimized

    def eventFilter(self, obj, evt):
        """Handle click events for the tabbar"""
        if obj == self.__tabBar:
            if evt.type() == QEvent.MouseButtonPress:
                pos = evt.pos()

                index = None
                for index in range(self.__tabBar.count()):
                    if self.__tabBar.tabRect(index).contains(pos):
                        break

                if index == self.__tabBar.currentIndex():
                    if self.isMinimized():
                        self.expand()
                    else:
                        self.shrink()
                    return True

                if self.isMinimized():
                    if self.isTabEnabled(index):
                        self.expand()

        return QWidget.eventFilter(self, obj, evt)

    @staticmethod
    def __toVariant(name, priority):
        """A tab stores its name and priority"""
        if priority is None:
            return ':' + name
        return str(priority) + ':' + name

    @staticmethod
    def __fromVariant(data):
        """A tab stores its name and priority"""
        val = data.split(':')
        if val[0] == '':
            return val[1], None
        return val[1], int(val[0])

    def __appendTab(self, widget, icon, label, name, priority):
        """Appends a new widget to the end"""
        self.__tabBar.addTab(icon, label)
        self.__stackedWidget.addWidget(widget)
        self.__tabBar.setTabData(self.count - 1,
                                 self.__toVariant(name, priority))

    def __pickInsertIndex(self, priority):
        """Picks the tab insert index in accordance to the priority"""
        for index in range(self.__tabBar.count()):
            data = self.__tabBar.tabData(index)
            _, tabPriority = self.__fromVariant(data)
            if tabPriority is None:
                return index
            if priority < tabPriority:
                return index
        return None

    def addTab(self, widget, icon, label, name, priority):
        """Add a tab to the sidebar"""
        if priority is None:
            # Appending to the end
            self.__appendTab(widget, icon, label, name, priority)
        else:
            # Pick the index in accordance to the priority
            index = self.__pickInsertIndex(priority)
            if index is None:
                self.__appendTab(widget, icon, label, name, priority)
            else:
                self.__tabBar.insertTab(index, icon, label)
                self.__stackedWidget.insertWidget(index, widget)
                self.__tabBar.setTabData(index,
                                         self.__toVariant(name, priority))

    def clear(self):
        """Remove all tabs"""
        while self.count > 0:
            self.removeTab(0)

    def prevTab(self):
        """Show the previous tab"""
        index = self.currentIndex() - 1
        if index < 0:
            index = self.count - 1

        self.setCurrentIndex(index)
        self.currentWidget().setFocus()

    def nextTab(self):
        """Show the next tab"""
        index = self.currentIndex() + 1
        if index >= self.count:
            index = 0

        self.setCurrentIndex(index)
        self.currentWidget().setFocus()

    @property
    def count(self):
        """Provides the number of tabs"""
        return self.__tabBar.count()

    def currentIndex(self):
        """Provides the index of the current tab"""
        return self.__stackedWidget.currentIndex()

    def currentWidget(self):
        """Provide a reference to the current widget"""
        return self.__stackedWidget.currentWidget()

    def currentTabName(self):
        """Provides the name of the current tab"""
        return self.getTabName(self.currentIndex())

    def __getWidgetIndex(self, indexOrNameOrWidget):
        """Provides the widget index via the provided index, name or widget"""
        if indexOrNameOrWidget is None:
            return None
        if isinstance(indexOrNameOrWidget, int):
            if indexOrNameOrWidget >= self.count:
                return None
            return indexOrNameOrWidget
        if isinstance(indexOrNameOrWidget, str):
            for index in range(self.__tabBar.count()):
                data = self.__tabBar.tabData(index)
                tabName, _ = self.__fromVariant(data)
                if tabName == indexOrNameOrWidget:
                    return index
            return None
        return self.__stackedWidget.indexOf(indexOrNameOrWidget)

    def getTabName(self, indexOrWidget):
        """Provides the tab name by index or widget"""
        index = self.__getWidgetIndex(indexOrWidget)
        if index is not None:
            data = self.__tabBar.tabData(index)
            tabName, _ = self.__fromVariant(data)
            return tabName
        return None

    def updateTabName(self, indexOrNameOrWidget, newName):
        """Updates the tab name"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            data = self.__tabBar.tabData(index)
            _, priority = self.__fromVariant(data)
            self.__tabBar.setTabData(index,
                                     self.__toVariant(newName, priority))

    def setCurrentTab(self, indexOrNameOrWidget):
        """Sets the current widget approprietly"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            self.__tabBar.setCurrentIndex(index)
            self.__stackedWidget.setCurrentIndex(index)
            if self.isMinimized():
                self.expand()

    def removeTab(self, indexOrNameOrWidget):
        """Remove a tab"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            self.__stackedWidget.removeWidget(
                self.__stackedWidget.widget(index))
            self.__tabBar.removeTab(index)

    def isTabEnabled(self, indexOrNameOrWidget):
        """Check if the tab is enabled"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            return self.__tabBar.isTabEnabled(index)
        return False

    def setTabEnabled(self, indexOrNameOrWidget, enabled):
        """Set the enabled state of the tab"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            self.__tabBar.setTabEnabled(index, enabled)

    def orientation(self):
        """Provides the orientation of the sidebar"""
        return self.__orientation

    def setOrientation(self, orient):
        """Set the orientation of the sidebar"""
        if orient == SideBar.North:
            self.__tabBar.setShape(QTabBar.RoundedNorth)
            self.barLayout.setDirection(QBoxLayout.LeftToRight)
            self.layout.setDirection(QBoxLayout.TopToBottom)
            self.layout.setAlignment(self.barLayout, Qt.AlignLeft)
        elif orient == SideBar.East:
            self.__tabBar.setShape(QTabBar.RoundedEast)
            self.barLayout.setDirection(QBoxLayout.TopToBottom)
            self.layout.setDirection(QBoxLayout.RightToLeft)
            self.layout.setAlignment(self.barLayout, Qt.AlignTop)
        elif orient == SideBar.South:
            self.__tabBar.setShape(QTabBar.RoundedSouth)
            self.barLayout.setDirection(QBoxLayout.LeftToRight)
            self.layout.setDirection(QBoxLayout.BottomToTop)
            self.layout.setAlignment(self.barLayout, Qt.AlignLeft)
        else:
            # default
            orient = SideBar.West
            self.__tabBar.setShape(QTabBar.RoundedWest)
            self.barLayout.setDirection(QBoxLayout.TopToBottom)
            self.layout.setDirection(QBoxLayout.LeftToRight)
            self.layout.setAlignment(self.barLayout, Qt.AlignTop)
        self.__orientation = orient

    def tabIcon(self, indexOrNameOrWidget):
        """Provide the icon of the tab"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            return self.__tabBar.tabIcon(index)
        return None

    def setTabIcon(self, indexOrNameOrWidget, icon):
        """Set the icon of the tab"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            self.__tabBar.setTabIcon(index, icon)

    def tabText(self, indexOrNameOrWidget):
        """Provide the text of the tab"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            return self.__tabBar.tabText(index)
        return None

    def tabButton(self, indexOrNameOrWidget, position):
        """Provide the button of the tab"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            return self.__tabBar.tabButton(index, position)
        return None

    def setTabText(self, indexOrNameOrWidget, text):
        """Set the text of the tab"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            self.__tabBar.setTabText(index, text)

    def tabToolTip(self, indexOrNameOrWidget):
        """Provide the tooltip text of the tab"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            return self.__tabBar.tabToolTip(index)
        return None

    def setTabToolTip(self, indexOrNameOrWidget, tip):
        """Set the tooltip text of the tab"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            self.__tabBar.setTabToolTip(index, tip)

    def tabWhatsThis(self, indexOrNameOrWidget):
        """Provide the WhatsThis text of the tab"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            return self.__tabBar.tabWhatsThis(index)
        return None

    def setTabWhatsThis(self, indexOrNameOrWidget, text):
        """Set the WhatsThis text for the tab"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            self.__tabBar.setTabWhatsThis(index, text)

    def widget(self, indexOrNameOrWidget):
        """Provides the reference to the widget: QWidget or None"""
        index = self.__getWidgetIndex(indexOrNameOrWidget)
        if index is not None:
            return self.__stackedWidget.widget(index)
        return None

    def setTabsClosable(self, closable):
        """Sets the tabs closable"""
        self.__tabBar.setTabsClosable(closable)

    def __onCloseRequest(self, index):
        """Re-emits the close request signal"""
        self.sigTabCloseRequested.emit(index)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Color button"""

from .qt import pyqtSignal, QPushButton, QColorDialog


class ColorButton(QPushButton):

    """A button which uses the selected color as its background"""

    sigColorChanged = pyqtSignal()

    def __init__(self, *args, **kwargs):
        QPushButton.__init__(self, *args, **kwargs)

        self.__color = None
        self.pressed.connect(self.onColorPicker)

    def setColor(self, color):
        """Sets the button color"""
        if color != self.__color:
            self.__color = color
            if self.__color:
                self.setStyleSheet('background-color: ' + self.__color.name())
            else:
                self.setStyleSheet('')
            self.sigColorChanged.emit()

    def color(self):
        """Provides the current color"""
        return self.__color

    def onColorPicker(self):
        """Brings up the standard color picking dialog"""
        dlg = QColorDialog(self.parent())
        dlg.setOptions(QColorDialog.DontUseNativeDialog)
        if self.__color:
            dlg.setCurrentColor(self.__color)

        if dlg.exec_():
            self.setColor(dlg.currentColor())



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2020  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Spacers to be used on the toolbars"""


from .qt import QWidget, QSizePolicy


class ToolBarHSpacer(QWidget):

    """Horizontal toolbar spacer"""

    def __init__(self, parent, width, name=None):
        QWidget.__init__(self, parent)
        self.setFixedWidth(width)
        if name is not None:
            self.setObjectName(name)
        self.setStyleSheet('background: transparent')


class ToolBarVSpacer(QWidget):

    """Vertical toolbar spacer"""

    def __init__(self, parent, height, name=None):
        QWidget.__init__(self, parent)
        self.setFixedHeight(height)
        if name is not None:
            self.setObjectName(name)
        self.setStyleSheet('background: transparent')


class ToolBarExpandingSpacer(QWidget):

    """Expanding toolbar spacer"""

    def __init__(self, parent, name=None):
        QWidget.__init__(self, parent)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        if name is not None:
            self.setObjectName(name)
        self.setStyleSheet('background: transparent')




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2015-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Virtual canvas to represent a control flow.

The basic idea is to split the canvas into cells and each cell could
be whether a vacant or filled with a certain graphics element.

At the very beginning the canvas is empty and can grow left, right or down
when a new element needs to be inserted.

The whole canvas is split into independent sections. The growing in one section
does not affect all the other sections.
"""

from ui.qt import QColor, QPen, QBrush
from cdmcfparser import (CODEBLOCK_FRAGMENT, FUNCTION_FRAGMENT, CLASS_FRAGMENT,
                         BREAK_FRAGMENT, CONTINUE_FRAGMENT, RETURN_FRAGMENT,
                         RAISE_FRAGMENT, ASSERT_FRAGMENT, SYSEXIT_FRAGMENT,
                         IMPORT_FRAGMENT, COMMENT_FRAGMENT,
                         WHILE_FRAGMENT, FOR_FRAGMENT, IF_FRAGMENT,
                         WITH_FRAGMENT, TRY_FRAGMENT, CML_COMMENT_FRAGMENT)
from .cml import CMLVersion, CMLsw, CMLgb, CMLge, CMLdoc
from .cellelement import CellElement
from .items import (CodeBlockCell, ReturnCell, RaiseCell, AssertCell,
                    SysexitCell, ImportCell,  IfCell)
from .minimizeditems import (MinimizedExceptCell,
                             MinimizedIndependentCommentCell,
                             MinimizedLeadingCommentCell,
                             MinimizedAboveCommentCell,
                             MinimizedSideCommentCell)
from .auxitems import ConnectorCell, VacantCell, VSpacerCell, Rectangle
from .loopjumpitems import BreakCell, ContinueCell
from .scopeitems import (ScopeCellElement, FileScopeCell, FunctionScopeCell,
                         ClassScopeCell, ForScopeCell, WhileScopeCell,
                         TryScopeCell, WithScopeCell, DecoratorScopeCell,
                         ExceptScopeCell, FinallyScopeCell,
                         ForElseScopeCell, WhileElseScopeCell,
                         TryElseScopeCell, ElseScopeCell,
                         ScopeHSideEdge, ScopeVSideEdge, ScopeSpacer)
from .commentitems import (AboveCommentCell, LeadingCommentCell,
                           SideCommentCell, IndependentCommentCell)
from .groupitems import (EmptyGroup, OpenedGroupBegin, OpenedGroupEnd,
                         CollapsedGroup, HGroupSpacerCell)
from .docitems import IndependentDocCell, LeadingDocCell, AboveDocCell
from .routines import getDocComment


CONN_N_S = [(ConnectorCell.NORTH, ConnectorCell.SOUTH)]
CONN_W_E = [(ConnectorCell.WEST, ConnectorCell.EAST)]
CONN_E_W = [(ConnectorCell.EAST, ConnectorCell.WEST)]
CONN_N_W = [(ConnectorCell.NORTH, ConnectorCell.WEST)]
CONN_W_S = [(ConnectorCell.WEST, ConnectorCell.SOUTH)]
CONN_E_S = [(ConnectorCell.EAST, ConnectorCell.SOUTH)]
CONN_N_C = [(ConnectorCell.NORTH, ConnectorCell.CENTER)]

_scopeToClass = {
    CellElement.NO_SCOPE: None,
    CellElement.FILE_SCOPE: FileScopeCell,
    CellElement.FUNC_SCOPE: FunctionScopeCell,
    CellElement.CLASS_SCOPE: ClassScopeCell,
    CellElement.FOR_SCOPE: ForScopeCell,
    CellElement.WHILE_SCOPE: WhileScopeCell,
    CellElement.TRY_SCOPE: TryScopeCell,
    CellElement.WITH_SCOPE: WithScopeCell,
    CellElement.DECOR_SCOPE: DecoratorScopeCell,
    CellElement.FOR_ELSE_SCOPE: ForElseScopeCell,
    CellElement.WHILE_ELSE_SCOPE: WhileElseScopeCell,
    CellElement.TRY_ELSE_SCOPE: TryElseScopeCell,
    CellElement.EXCEPT_SCOPE: ExceptScopeCell,
    CellElement.FINALLY_SCOPE: FinallyScopeCell}

_fragmentKindToCellClass = {
    CODEBLOCK_FRAGMENT: CodeBlockCell,
    BREAK_FRAGMENT: BreakCell,
    CONTINUE_FRAGMENT: ContinueCell,
    RETURN_FRAGMENT: ReturnCell,
    RAISE_FRAGMENT: RaiseCell,
    ASSERT_FRAGMENT: AssertCell,
    SYSEXIT_FRAGMENT: SysexitCell,
    IMPORT_FRAGMENT: ImportCell}

_scopeToName = {
    CellElement.FILE_SCOPE: "",
    CellElement.FOR_SCOPE: "for",
    CellElement.WHILE_SCOPE: "while",
    CellElement.TRY_SCOPE: "try",
    CellElement.WITH_SCOPE: "with",
    CellElement.EXCEPT_SCOPE: "except",
    CellElement.FINALLY_SCOPE: "finally"}


_terminalCellTypes = (
    CellElement.BREAK,
    CellElement.CONTINUE,
    CellElement.RETURN,
    CellElement.RAISE,
    CellElement.SYSEXIT)


class VirtualCanvas:

    """Holds the control flow representation"""

    def __init__(self, settings, xAddr, yAddr,
                 validGroups, collapsedGroups, parent):
        self.kind = CellElement.VCANVAS

        # the item instances from items.py or other virtual canvases
        self.cells = []

        # Reference to the upper level canvas or None for the most upper canvas
        self.canvas = parent
        self.editor = None

        self.settings = settings
        self.addr = [xAddr, yAddr]

        # Layout support
        self.__currentCF = None
        self.__currentScopeClass = None
        self.isNoScope = False

        # Rendering support
        self.width = 0
        self.height = 0
        self.minWidth = 0
        self.minHeight = 0
        self.linesInHeader = 0

        # inclusive regions of rows with columns which affect each other width
        self.dependentRegions = []

        self.tailComment = False

        # Painting support
        self.baseX = 0
        self.baseY = 0
        self.scopeRectangle = None

        # Groups support
        self.__validGroups = validGroups
        self.__validGroupBeginLines = [item[1] for item in validGroups]
        self.__validGroupEndLines = [item[2] for item in validGroups]
        self.__validGroupLines = self.__validGroupBeginLines + \
                                 self.__validGroupEndLines
        self.__collapsedGroups = collapsedGroups
        self.__groupStack = []  # [item, row, column]
        self.maxLocalOpenGroupDepth = 0
        self.maxGlobalOpenGroupDepth = 0

        # Supports the second stage of open group layout. If True => this
        # nested layout needs to be considered on the upper levels with how
        # the spacing is inserted for the open groups
        self.isIfBelowLayout = False
        self.isOuterIfLayout = False

    def cleanup(self):
        """Cleans up the references etc"""
        self.canvas = None
        self.editor = None
        self.settings = None
        self.addr = None
        self.__currentCF = None
        self.__currentScopeClass = None
        self.scopeRectangle = None
        self.__validGroups = None
        self.__validGroupBeginLines = None
        self.__validGroupEndLines = None
        self.__validGroupLines = None
        self.__collapsedGroups = None
        self.__groupStack = None

        for row in self.cells:
            if row:
                for cell in row:
                    if isinstance(cell, VirtualCanvas):
                        cell.cleanup()

    @staticmethod
    def scopedItem():
        """The vcanvas is not a scoped item"""
        return False

    @staticmethod
    def isComment():
        """The vcanvas is not a comment"""
        return False

    @staticmethod
    def isCMLDoc():
        """The vcanvas is not a CML doc"""
        return False

    def getScopeName(self):
        """Provides the name of the scope drawn on the canvas"""
        for row in self.cells:
            for cell in row:
                if cell.kind in _scopeToName:
                    return _scopeToName[cell.kind]
                if cell.kind == CellElement.FUNC_SCOPE:
                    return 'def ' + cell.ref.name.getContent() + '()'
                if cell.kind == CellElement.CLASS_SCOPE:
                    return 'class ' + cell.ref.name.getContent()
                if cell.kind == CellElement.DECOR_SCOPE:
                    return '@' + cell.ref.name.getContent()
                if cell.kind == CellElement.ELSE_SCOPE:
                    if cell.statement == ElseScopeCell.TRY_STATEMENT:
                        return 'try-else'
                    if cell.statement == ElseScopeCell.FOR_STATEMENT:
                        return 'for-else'
                    if cell.statement == ElseScopeCell.WHILE_STATEMENT:
                        return 'while-else'
                    # The 'statement' is set only for the TOP_LEFT item of the
                    # ELSE_SCOPE so it must be found here because we start from
                    # the top left corner
                    return '???'
        return None

    def __str__(self):
        """Rather debug support"""
        val = '\nRows: ' + str(len(self.cells))
        count = 0
        for row in self.cells:
            val += '\nRow ' + str(count) + ' (' + str(len(row)) + ' items): [ '
            for item in row:
                val += str(item) + ', '
            val += ']'
            count += 1
        return val

    def __isTerminalCell(self, row, column):
        """Tells if a cell is terminal.

        i.e. no need to continue the control flow line.
        """
        try:
            # The previous row could be a group end...
            while self.cells[row][-1].kind in [CellElement.OPENED_GROUP_END,
                                               CellElement.OPENED_GROUP_BEGIN]:
                row -= 1

            cell = self.cells[row][column]
            if cell.kind == CellElement.COLLAPSED_GROUP:
                return self.__isCollapsedGroupTerminal(cell)

            if cell.kind == CellElement.VCANVAS:
                # If it is a scope then the last item in the scope should be
                # checked
                cells = cell.cells
                rowIndex = len(cells) - 1
                while cells[rowIndex][-1].kind in [CellElement.OPENED_GROUP_END,
                                                   CellElement.OPENED_GROUP_BEGIN]:
                    rowIndex -= 1
                cell = cell.cells[rowIndex][0]
            if cell.kind == CellElement.CONNECTOR:
                if cell.connections == CONN_N_C:
                    # On some smart zoom levels the primitives are replaced
                    # with connectors. The terminal cells are replaced with a
                    # half a connector from NORTH to CENTER
                    return True
            return cell.kind in _terminalCellTypes
        except:
            return False

    def __isCollapsedGroupTerminal(self, group):
        """Tells if the collapsed group has the last item terminal"""
        if group.nestedRefs[-1].CODE in [BREAK_FRAGMENT,
                                         CONTINUE_FRAGMENT,
                                         RETURN_FRAGMENT,
                                         RAISE_FRAGMENT,
                                         SYSEXIT_FRAGMENT]:
            return True

    def __isVacantCell(self, row, column):
        """Tells if a cell is a vacant one"""
        try:
            return self.cells[row][column].kind == CellElement.VACANT
        except:
            return True

    def __allocateCell(self, row, column, needScopeEdge=True):
        """Allocates a cell as Vacant if it is not available yet.

        Can only allocate bottom and right growing cells.
        """
        lastIndex = len(self.cells) - 1
        while lastIndex < row:
            self.cells.append([])
            lastIndex += 1
            if needScopeEdge:
                if self.__currentScopeClass:
                    self.cells[lastIndex].append(
                        ScopeHSideEdge(self.__currentCF, self, 0, lastIndex))
        lastIndex = len(self.cells[row]) - 1
        while lastIndex < column:
            self.cells[row].append(VacantCell(None, self, lastIndex, row))
            lastIndex += 1

    def __allocateAndSet(self, row, column, what):
        """Allocates a cell and sets it to the given value"""
        self.__allocateCell(row, column)
        self.cells[row][column] = what

    def __allocateScope(self, item, scopeType, row, column):
        """Allocates a scope for a suite"""
        canvas = VirtualCanvas(self.settings, column, row,
                               self.__validGroups, self.__collapsedGroups,
                               self)
        canvas.layout(item, scopeType)
        self.__allocateAndSet(row, column, canvas)

    def __allocateLeadingComment(self, item, row, column):
        """Allocates a leading comment if so"""
        if not self.settings.noComment:
            leadingDoc = getDocComment(item.leadingCMLComments)
            if leadingDoc:
                self.__allocateCell(row, column + 1)
                self.cells[row][column] = ConnectorCell(CONN_N_S, self,
                                                        column, row)
                self.cells[row][column + 1] = LeadingDocCell(item, leadingDoc,
                                                             self,
                                                             column + 1, row)
                row += 1

            if item.leadingComment:
                self.__allocateCell(row, column + 1)
                self.cells[row][column] = ConnectorCell(CONN_N_S, self,
                                                        column, row)

                self.cells[row][column + 1] = self.__createLeadingComment(
                    item, self, column + 1, row)
                row += 1
        return row

    def __needLoopCommentRow(self, item):
        """Tells # of rows to be reserved for comments/docs for the loops"""
        if self.settings.noComment:
            return 0, 0, []

        comments = []

        doc = getDocComment(item.leadingCMLComments)
        comments.append([doc, item.leadingComment])
        rows = 0
        if item.leadingComment:
            rows += 1
        if doc:
            rows += 1

        elseRows = 0
        if item.elsePart:
            elseDoc = getDocComment(item.elsePart.leadingCMLComments)
            comments.append([elseDoc, item.elsePart.leadingComment])

            if item.elsePart.leadingComment:
                elseRows += 1
            if elseDoc:
                elseRows += 1
        else:
            comments.append([None, None])
        return rows, elseRows, comments

    def __needTryCommentRow(self, item):
        """Tells if a row for comments need to be reserved"""
        if self.settings.noComment:
            return 0, []

        comments = []

        doc = getDocComment(item.leadingCMLComments)
        rows = 0
        if item.leadingComment:
            rows += 1
        if doc:
            rows += 1
        comments.append([rows, doc, item.leadingComment])

        if not self.settings.hideexcepts:
            for exceptPart in item.exceptParts:
                excDoc = getDocComment(exceptPart.leadingCMLComments)

                excRows = 0
                if exceptPart.leadingComment:
                    excRows += 1
                if excDoc:
                    excRows += 1
                comments.append([excRows, excDoc, exceptPart.leadingComment])

                rows = max(rows, excRows)
        return rows, comments

    def __checkLeadingCMLComments(self, leadingCMLComments):
        """Provides a list of group begins/ends as they are in the comments"""
        groups = []
        for comment in leadingCMLComments:
            if hasattr(comment, 'CODE'):
                if comment.CODE in [CMLgb.CODE, CMLge.CODE]:
                    itemFirstLine = comment.ref.parts[0].beginLine
                    if itemFirstLine in self.__validGroupLines:
                        groups.append(comment)
        return groups

    def __getGroups(self, item):
        """Provides a list of group begins and ends as they are in the item"""
        if type(item) == list:
            # this is a list of CML comments
            return self.__checkLeadingCMLComments(item)

        # Only valid groups are taken into account
        if item.kind == CML_COMMENT_FRAGMENT:
            if hasattr(item, 'ref'):
                # High level CML comment (low level are not interesting here)
                itemFirstLine = item.ref.parts[0].beginLine
                if itemFirstLine in self.__validGroupLines:
                    return [item]
            return []

        if item.kind == IF_FRAGMENT:
            return self.__checkLeadingCMLComments(
                item.parts[0].leadingCMLComments)

        # The item may have a leading CML comment which ends a valid group
        if not hasattr(item, 'leadingCMLComments'):
            return []
        return self.__checkLeadingCMLComments(item.leadingCMLComments)

    def __isGroupBegin(self, groupComment):
        """True if it is a valid group begin"""
        return groupComment.CODE == CMLgb.CODE

    def __onGroupBegin(self, item, groupComment, vacantRow, column):
        """Handles the group begin"""
        if self.__groupStack:
            if self.__groupStack[-1][0].kind == CellElement.COLLAPSED_GROUP:
                # We are still in a collapsed group
                return vacantRow

        groupId = groupComment.id
        if self.__isGroupCollapsed(groupId):
            newGroup = CollapsedGroup(item, groupComment, self,
                                      column, vacantRow)
        else:
            newGroup = OpenedGroupBegin(item, groupComment, self,
                                        column, vacantRow)
            newGroup.isTerminal = self.__isTerminalCell(vacantRow - 1, column)

        # allocate new cell, memo the group begin ref,
        # add the group to the stack and return a new vacant row
        self.__allocateAndSet(vacantRow, column, newGroup)
        self.__groupStack.append([newGroup, vacantRow, column])
        return vacantRow + 1

    def __onGroupEnd(self, item, groupComment, vacantRow, column):
        """Handles the group end"""
        # At least one group must be in the stack!
        currentGroup = self.__groupStack[-1][0]
        if currentGroup.getGroupId() != groupComment.id:
            # It is not a close of the current group
            return vacantRow

        groupBegin = self.__groupStack[-1][0]
        groupRow = self.__groupStack[-1][1]
        groupColumn = self.__groupStack[-1][2]

        # There are three cases here:
        # - end of a collapsed group
        # - end of an empty group
        # - end of an open group
        if not currentGroup.nestedRefs:
            # Empty group: replace the beginning of the group with an empty one
            emptyGroup = EmptyGroup(groupBegin.ref,
                                    groupBegin.groupBeginCMLRef, self,
                                    groupColumn, groupRow)
            self.cells[groupRow][groupColumn] = emptyGroup
        elif currentGroup.kind == CellElement.COLLAPSED_GROUP:
            # Collapsed group: the end of the group is memorized in the common
            # block after ifs
            pass
        else:
            # Opened group: insert a group end
            groupEnd = OpenedGroupEnd(item, groupBegin.groupBeginCMLRef,
                                      self, column, vacantRow)
            groupEnd.groupBeginRow = groupRow
            groupEnd.groupBeginColumn = groupColumn

            groupEnd.isTerminal = self.__isTerminalCell(vacantRow - 1, column)
            self.__allocateAndSet(vacantRow, column, groupEnd)

            self.cells[groupRow][groupColumn].groupEndRow = vacantRow
            self.cells[groupRow][groupColumn].groupEndColumn = column

            vacantRow += 1

        self.cells[groupRow][groupColumn].groupEndCMLRef = groupComment

        self.__groupStack.pop()
        return vacantRow

    def __isGroupCollapsed(self, groupId):
        """True if the group is collapsed"""
        if self.__collapsedGroups:
            return groupId in self.__collapsedGroups
        return False

    def __handleGroups(self, item, vacantRow, column):
        """Picks all the valid group begins and ends

        Decides what to do and returns True if it is a collapsed group and
        the further items need to be skipped
        """
        for groupComment in self.__getGroups(item):
            # The group begin/end comments are going to the nested refs as
            # separate items regardless from where they came. This lets to
            # detect empty groups easier.
            if self.__groupStack:
                currentGroupId = self.__groupStack[-1][0].getGroupId()
                if currentGroupId != groupComment.id:
                    # The 'if' is to avoid the end of the group to be added
                    # to the nested refs
                    self.__groupStack[-1][0].nestedRefs.append(groupComment)

            if self.__isGroupBegin(groupComment):
                vacantRow = self.__onGroupBegin(item, groupComment,
                                                vacantRow, column)
            else:
                vacantRow = self.__onGroupEnd(item, groupComment,
                                              vacantRow, column)

        # type(item) != list protects from dealing with a list of CML comments
        if self.__groupStack and type(item) != list:
            # We are in some kind of a group
            needToAdd = True
            if item.kind == CML_COMMENT_FRAGMENT:
                if hasattr(item, 'ref'):
                    # That's a high level comment
                    if item.CODE in [CMLgb.CODE, CMLge.CODE]:
                        if item.id == self.__groupStack[-1][0].getGroupId():
                            needToAdd = False
            if needToAdd:
                self.__groupStack[-1][0].nestedRefs.append(item)
            if self.__groupStack[-1][0].kind == CellElement.COLLAPSED_GROUP:
                return True, vacantRow

        # We are not in a group now, so process the item
        return False, vacantRow

    def __checkOpenGroupBefore(self, vacantRow, column):
        """Checks if the previous row is an open group end"""
        if vacantRow > 0:
            for cell in self.cells[vacantRow - 1]:
                if cell.kind == CellElement.OPENED_GROUP_END:
                    # Need to insert a connector or a spacer
                    if cell.isTerminal:
                        spacerColumn = column
                    else:
                        spacerColumn = column + 1
                        conn = ConnectorCell(CONN_N_S, self, column, vacantRow)
                        self.__allocateAndSet(vacantRow, column, conn)
                    spacer = VSpacerCell(None, self, spacerColumn, vacantRow)
                    self.__allocateAndSet(vacantRow, spacerColumn, spacer)

                    vacantRow += 1
                    return vacantRow
        return vacantRow

    def __createIndependentComment(self, ref, canvas, xPos, yPos):
        """Creates an independent comment"""
        if self.settings.hidecomments:
            return MinimizedIndependentCommentCell(ref, canvas, xPos, yPos)
        return IndependentCommentCell(ref, canvas, xPos, yPos)

    def __createLeadingComment(self, ref, canvas, xPos, yPos):
        """Creates a leading comment"""
        if self.settings.hidecomments:
            return MinimizedLeadingCommentCell(ref, canvas, xPos, yPos)
        return LeadingCommentCell(ref, canvas, xPos, yPos)

    def __createAboveComment(self, ref, canvas, xPos, yPos):
        """Creates an above comment"""
        if self.settings.hidecomments:
            return MinimizedAboveCommentCell(ref, canvas, xPos, yPos)
        return AboveCommentCell(ref, canvas, xPos, yPos)

    def __createSideComment(self, ref, canvas, xPos, yPos):
        """Creates a side comment"""
        if self.settings.hidecomments:
            return MinimizedSideCommentCell(ref, canvas, xPos, yPos)
        return SideCommentCell(ref, canvas, xPos, yPos)

    def __layoutCMLComment(self, item, vacantRow, column):
        """Lays out a CML comment"""
        if hasattr(item, 'ref'):
            # High level CML comment, low level are out of interest
            if item.CODE == CMLdoc.CODE:
                conn = ConnectorCell(CONN_N_S, self, column, vacantRow)
                doc = IndependentDocCell(item, self, column + 1, vacantRow)

                self.__allocateCell(vacantRow, column + 1)
                self.cells[vacantRow][column] = conn
                self.cells[vacantRow][column + 1] = doc
                vacantRow += 1
        return vacantRow

    def __layoutWith(self, item, vacantRow, column):
        """Lays out a with statement"""
        tempVacantRow = vacantRow
        vacantRow = self.__allocateLeadingComment(item, vacantRow, column)
        if tempVacantRow == vacantRow:
            # Nothing has been inserted, i.e. may need a spacer after
            # the end of an opened group
            vacantRow = self.__checkOpenGroupBefore(vacantRow, column)

        self.__allocateScope(item, CellElement.WITH_SCOPE, vacantRow, column)
        return vacantRow + 1

    def __layoutComment(self, item, vacantRow, column):
        """Lays out a comment"""
        conn = ConnectorCell(CONN_N_S, self, column, vacantRow)
        comment = self.__createIndependentComment(item, self, column + 1,
                                                  vacantRow)

        self.__allocateCell(vacantRow, column + 1)
        self.cells[vacantRow][column] = conn
        self.cells[vacantRow][column + 1] = comment
        return vacantRow + 1

    def __layoutDefClass(self, item, vacantRow, column):
        """Lays out a function or a class"""
        scopeCanvas = VirtualCanvas(self.settings, None, None,
                                    self.__validGroups, self.__collapsedGroups,
                                    self)
        scopeItem = item
        if item.kind == FUNCTION_FRAGMENT:
            scopeCanvas.layout(item, CellElement.FUNC_SCOPE)
        else:
            scopeCanvas.layout(item, CellElement.CLASS_SCOPE)

        if item.decorators and not self.settings.noDecor:
            for dec in reversed(item.decorators):
                # Create a decorator scope virtual canvas
                decScope = VirtualCanvas(self.settings, None, None,
                                         self.__validGroups,
                                         self.__collapsedGroups, self)
                decScopeRows = len(decScope.cells)

                needRows = 1
                leadingDoc = getDocComment(scopeItem.leadingCMLComments)
                if not self.settings.noComment:
                    if leadingDoc is not None:
                        needRows += 1
                    if scopeItem.leadingComment:
                        needRows += 1

                decScope.layout(dec, CellElement.DECOR_SCOPE, needRows)
                if needRows > 1:
                    # Need one or two comments
                    if leadingDoc is not None:
                        conn = ConnectorCell(CONN_N_S, decScope, 1, 3)
                        doc = LeadingDocCell(scopeItem, leadingDoc,
                                             decScope, 2, 3)

                        decScope.__allocateCell(3, 2)
                        decScope.cells[3][1] = conn
                        decScope.cells[3][2] = doc

                    if scopeItem.leadingComment:
                        rowAddr = needRows + 1
                        conn = ConnectorCell(CONN_N_S, decScope, 1, rowAddr)
                        comment = self.__createLeadingComment(
                            scopeItem, decScope, 2, rowAddr)

                        decScope.__allocateCell(rowAddr, 2)
                        decScope.cells[rowAddr][1] = conn
                        decScope.cells[rowAddr][2] = comment

                decScope.__allocateCell(decScopeRows - 2, 1)

                # Fix the parent
                scopeCanvas.parent = decScope
                scopeCanvas.canvas = decScope
                # Set the decorator content
                decScope.cells[decScopeRows - 2][1] = scopeCanvas
                scopeCanvas.addr = [1, decScopeRows - 2]
                # Set the current content scope
                scopeCanvas = decScope
                scopeItem = dec

        tempVacantRow = vacantRow
        vacantRow = self.__allocateLeadingComment(scopeItem, vacantRow, column)
        if tempVacantRow == vacantRow:
            # Nothing has been inserted, i.e. may need a spacer after
            # the end of an opened group
            vacantRow = self.__checkOpenGroupBefore(vacantRow, column)

        # Update the scope canvas parent and address
        scopeCanvas.parent = self
        scopeCanvas.addr = [column, vacantRow]
        self.__allocateAndSet(vacantRow, column, scopeCanvas)
        return vacantRow + 1

    def __layoutLoop(self, item, vacantRow, column):
        """Lays out a loop"""
        targetScope = CellElement.WHILE_SCOPE
        if item.kind == FOR_FRAGMENT:
            targetScope = CellElement.FOR_SCOPE

        loopRegionBegin = vacantRow
        mainRows, elseRows, aboveItems = self.__needLoopCommentRow(item)
        maxRows = max(mainRows, elseRows)
        if maxRows > 0:
            # Main part
            cRow = vacantRow + (maxRows - mainRows)

            tempVacant = vacantRow
            while mainRows < maxRows:
                conn = ConnectorCell(CONN_N_S, self, column, tempVacant)
                self.__allocateCell(tempVacant, column)
                self.cells[tempVacant][column] = conn
                tempVacant += 1
                mainRows += 1

            if aboveItems[0][0]:
                doc = AboveDocCell(item, aboveItems[0][0], self, column, cRow)
                doc.needConnector = True
                self.__allocateAndSet(cRow, column, doc)
                cRow += 1
            if item.leadingComment:
                comment = self.__createAboveComment(item, self, column, cRow)
                comment.needConnector = True
                self.__allocateAndSet(cRow, column, comment)

            if elseRows > 0:
                cRow = vacantRow + (maxRows - elseRows)

                if aboveItems[1][0]:
                    doc = AboveDocCell(item.elsePart, aboveItems[1][0], self,
                                       column + 1, cRow)
                    self.__allocateAndSet(cRow, column + 1, doc)
                    cRow += 1
                if aboveItems[1][1]:
                    comment = self.__createAboveComment(item.elsePart, self,
                                                        column + 1, cRow)
                    comment.needConnector = aboveItems[1][0] is not None
                    self.__allocateAndSet(cRow, column + 1, comment)

                self.dependentRegions.append((loopRegionBegin,
                                              vacantRow + maxRows))

            vacantRow += maxRows
        else:
            vacantRow = self.__checkOpenGroupBefore(vacantRow, column)

        self.__allocateScope(item, targetScope, vacantRow, column)
        if item.elsePart:
            if item.kind == FOR_FRAGMENT:
                self.__allocateScope(item.elsePart, CellElement.FOR_ELSE_SCOPE,
                                     vacantRow, column + 1)
            else:
                self.__allocateScope(item.elsePart,
                                     CellElement.WHILE_ELSE_SCOPE,
                                     vacantRow, column + 1)
            self.cells[vacantRow][column + 1].setLeaderRef(item)
        return vacantRow + 1

    def __layoutTry(self, item, vacantRow, column):
        """Lays out a try statement"""
        tryRegionBegin = vacantRow

        maxRows, aboveItems = self.__needTryCommentRow(item)
        if maxRows > 0:
            # Main part
            mainRows = aboveItems[0][0]
            cRow = vacantRow + (maxRows - mainRows)

            tempVacant = vacantRow
            while mainRows < maxRows:
                conn = ConnectorCell(CONN_N_S, self, column, tempVacant)

                self.__allocateCell(tempVacant, column)
                self.cells[tempVacant][column] = conn
                tempVacant += 1
                mainRows += 1

            if aboveItems[0][1]:
                doc = AboveDocCell(item, aboveItems[0][1], self, column, cRow)
                doc.needConnector = True
                self.__allocateAndSet(cRow, column, doc)
                cRow += 1
            if item.leadingComment:
                comment = self.__createAboveComment(item, self, column, cRow)
                comment.needConnector = True
                self.__allocateAndSet(cRow, column, comment)

            if item.exceptParts and not self.settings.hideexcepts:
                self.dependentRegions.append((tryRegionBegin,
                                              vacantRow + maxRows))


        self.__allocateScope(item, CellElement.TRY_SCOPE,
                             vacantRow + maxRows, column)

        if self.settings.hideexcepts:
            if item.exceptParts:
                miniExcept = MinimizedExceptCell(item, self, column + 1,
                                                 vacantRow + maxRows)
                self.__allocateAndSet(vacantRow + maxRows, column + 1,
                                      miniExcept)
        else:
            nextColumn = column + 1
            exceptIndex = 1
            for exceptPart in item.exceptParts:
                if maxRows > 0:
                    excRows = aboveItems[exceptIndex][0]
                    cRow = vacantRow + (maxRows - excRows)

                    if aboveItems[exceptIndex][1]:
                        doc = AboveDocCell(exceptPart,
                                           aboveItems[exceptIndex][1],
                                           self, nextColumn, cRow)
                        self.__allocateAndSet(cRow, nextColumn, doc)
                        cRow += 1
                    if aboveItems[exceptIndex][2]:
                        comment = self.__createAboveComment(exceptPart, self,
                                                            nextColumn, cRow)
                        comment.needConnector = \
                            aboveItems[exceptIndex][1] is not None
                        self.__allocateAndSet(cRow, nextColumn, comment)

                self.__allocateScope(exceptPart,
                                     CellElement.EXCEPT_SCOPE,
                                     vacantRow + maxRows, nextColumn)
                self.cells[vacantRow + maxRows][nextColumn].setLeaderRef(item)

                exceptIndex += 1
                nextColumn += 1

        vacantRow += maxRows

        # The else part goes below
        if item.elsePart:
            vacantRow += 1
            vacantRow = self.__allocateLeadingComment(item.elsePart,
                                                      vacantRow, column)
            self.__allocateScope(item.elsePart, CellElement.TRY_ELSE_SCOPE,
                                 vacantRow, column)
            self.cells[vacantRow][column].setLeaderRef(item)
        # The finally part is located below
        if item.finallyPart:
            vacantRow += 1
            vacantRow = self.__allocateLeadingComment(item.finallyPart,
                                                      vacantRow, column)
            self.__allocateScope(item.finallyPart, CellElement.FINALLY_SCOPE,
                                 vacantRow, column)
            self.cells[vacantRow][column].setLeaderRef(item)
        return vacantRow + 1

    def __layoutIf(self, item, vacantRow, column):
        """Lays out an if statement"""
        lastNonElseIndex = len(item.parts) - 1
        for index in range(len(item.parts)):
            if item.parts[index].condition is None:
                lastNonElseIndex = index - 1
                break

        canvas = VirtualCanvas(self.settings, 0, 0, self.__validGroups,
                               self.__collapsedGroups, self)
        canvas.isNoScope = True
        canvas.isIfBelowLayout = True
        canvas.isOuterIfLayout = True

        if lastNonElseIndex == len(item.parts) - 1:
            # There is no else
            canvas.layoutIfBranch(item.parts[lastNonElseIndex], None)
        else:
            canvas.layoutIfBranch(item.parts[lastNonElseIndex],
                                  item.parts[lastNonElseIndex + 1])

        index = lastNonElseIndex - 1
        while index >= 0:
            tempCanvas = VirtualCanvas(self.settings, 0, 0, self.__validGroups,
                                       self.__collapsedGroups, self)
            tempCanvas.isNoScope = True
            tempCanvas.layoutIfBranch(item.parts[index], canvas)
            canvas = tempCanvas
            index -= 1

        self.__allocateAndSet(vacantRow, 1, canvas)
        return vacantRow + 1

    def layoutSuite(self, vacantRow, suite,
                    scopeKind=None, cflow=None, column=1,
                    leadingCMLComments=None):
        """Does a single suite layout"""
        if scopeKind:
            self.__currentCF = cflow
            self.__currentScopeClass = _scopeToClass[scopeKind]

        skipItem = False
        if not self.settings.noGroup:
            if leadingCMLComments:
                skipItem, vacantRow = self.__handleGroups(leadingCMLComments,
                                                          vacantRow, column)

        for item in suite:

            if not self.settings.noGroup:
                skipItem, vacantRow = self.__handleGroups(item,
                                                          vacantRow, column)
                if skipItem:
                    continue

            if item.kind == CML_COMMENT_FRAGMENT:
                # CML comments are not shown on the diagram
                if not self.settings.noComment:
                    vacantRow = self.__layoutCMLComment(item, vacantRow, column)
                continue

            if item.kind == COMMENT_FRAGMENT:
                if not self.settings.noComment:
                    vacantRow = self.__layoutComment(item, vacantRow, column)
                continue

            if item.kind == WITH_FRAGMENT:
                if not self.settings.noWith:
                    vacantRow = self.__layoutWith(item, vacantRow, column)
                continue

            if item.kind in [FUNCTION_FRAGMENT, CLASS_FRAGMENT]:
                vacantRow = self.__layoutDefClass(item, vacantRow, column)
                continue

            if item.kind in [WHILE_FRAGMENT, FOR_FRAGMENT]:
                if item.kind == FOR_FRAGMENT and self.settings.noFor:
                    continue
                if item.kind == WHILE_FRAGMENT and self.settings.noWhile:
                    continue
                vacantRow = self.__layoutLoop(item, vacantRow, column)
                continue

            if item.kind == TRY_FRAGMENT:
                if not self.settings.noTry:
                    vacantRow = self.__layoutTry(item, vacantRow, column)
                continue

            if item.kind == IF_FRAGMENT:
                if not self.settings.noIf:
                    vacantRow = self.__layoutIf(item, vacantRow, column)
                continue

            # Below are the single cell fragments possibly with comments
            if item.kind == CODEBLOCK_FRAGMENT and self.settings.noBlock:
                self.__allocateCell(vacantRow, column)
                self.cells[vacantRow][column] = ConnectorCell(CONN_N_S, self,
                                                              column, vacantRow)
                vacantRow += 1
                continue
            if item.kind == IMPORT_FRAGMENT and self.settings.noImport:
                self.__allocateCell(vacantRow, column)
                self.cells[vacantRow][column] = ConnectorCell(CONN_N_S, self,
                                                              column, vacantRow)
                vacantRow += 1
                continue
            if item.kind == BREAK_FRAGMENT and self.settings.noBreak:
                self.__allocateCell(vacantRow, column)
                self.cells[vacantRow][column] = ConnectorCell(CONN_N_C, self,
                                                              column, vacantRow)
                vacantRow += 1
                continue
            if item.kind == CONTINUE_FRAGMENT and self.settings.noContinue:
                self.__allocateCell(vacantRow, column)
                self.cells[vacantRow][column] = ConnectorCell(CONN_N_C, self,
                                                              column, vacantRow)
                vacantRow += 1
                continue
            if item.kind == RETURN_FRAGMENT and self.settings.noReturn:
                self.__allocateCell(vacantRow, column)
                self.cells[vacantRow][column] = ConnectorCell(CONN_N_C, self,
                                                              column, vacantRow)
                vacantRow += 1
                continue
            if item.kind == RAISE_FRAGMENT and self.settings.noRaise:
                self.__allocateCell(vacantRow, column)
                self.cells[vacantRow][column] = ConnectorCell(CONN_N_C, self,
                                                              column, vacantRow)
                vacantRow += 1
                continue
            if item.kind == ASSERT_FRAGMENT and self.settings.noAssert:
                self.__allocateCell(vacantRow, column)
                self.cells[vacantRow][column] = ConnectorCell(CONN_N_S, self,
                                                              column, vacantRow)
                vacantRow += 1
                continue
            if item.kind == SYSEXIT_FRAGMENT and self.settings.noSysExit:
                self.__allocateCell(vacantRow, column)
                self.cells[vacantRow][column] = ConnectorCell(CONN_N_C, self,
                                                              column, vacantRow)
                vacantRow += 1
                continue

            cellClass = _fragmentKindToCellClass[item.kind]
            vacantRow = self.__allocateLeadingComment(item, vacantRow, column)
            self.__allocateAndSet(vacantRow, column,
                                  cellClass(item, self, column, vacantRow))

            if item.sideComment and not self.settings.noComment:
                comment = self.__createSideComment(item, self,
                                                   column + 1, vacantRow)
                self.__allocateAndSet(vacantRow, column + 1, comment)
            vacantRow += 1

            # end of for loop

        if scopeKind:
            self.__currentCF = None

        return vacantRow

    def layoutIfBranch(self, yBranch, nBranch):
        """Used in 'if' statements processing"""
        # It is always called when a layout is empty
        vacantRow = self.__allocateLeadingComment(yBranch, 0, 0)
        self.__allocateAndSet(vacantRow, 0,
                              IfCell(yBranch, self, 0, vacantRow))

        # Exclude the leading comments and docs from the dependent region
        dependentRegionBegin = vacantRow

        topConnector = ConnectorCell(CONN_W_S, self, 1, vacantRow)
        topConnector.subKind = ConnectorCell.TOP_IF
        self.__allocateAndSet(vacantRow, 1, topConnector)

        if yBranch.sideComment and not self.settings.noComment:
            comment = self.__createSideComment(yBranch, self, 2, vacantRow)
            self.__allocateAndSet(vacantRow, 2, comment)
        vacantRow += 1

        # Test if there is a switch of the branches
        yBelow = CMLVersion.find(yBranch.leadingCMLComments, CMLsw) is not None

        # Allocate the YES branch
        if yBelow:
            branchLayout = VirtualCanvas(self.settings, 0, vacantRow,
                                         self.__validGroups,
                                         self.__collapsedGroups, self)
            branchLayout.isIfBelowLayout = True
        else:
            branchLayout = VirtualCanvas(self.settings, 1, vacantRow,
                                         self.__validGroups,
                                         self.__collapsedGroups, self)
        branchLayout.isNoScope = True
        branchLayout.layoutSuite(0, yBranch.suite,
                                 CellElement.NO_SCOPE, None, 0)

        if yBelow:
            self.__allocateAndSet(vacantRow, 0, branchLayout)
        else:
            self.__allocateAndSet(vacantRow, 1, branchLayout)

        # nBranch could be: None: for absent of else
        #                   ifPart: present else
        #                   vcanvas: other elif
        if nBranch is None:
            if yBelow:
                self.__allocateAndSet(vacantRow, 1,
                                      ConnectorCell(CONN_N_S,
                                                    self, 1, vacantRow))
                vacantRow += 1
                bottomConnector = ConnectorCell(CONN_N_W, self, 1, vacantRow)
                bottomConnector.subKind = ConnectorCell.BOTTOM_IF
                self.__allocateAndSet(vacantRow, 1, bottomConnector)
                if self.__isTerminalCell(vacantRow - 1, 0) or \
                   self.__isVacantCell(vacantRow - 1, 0):
                    self.__allocateAndSet(vacantRow, 0,
                                          ConnectorCell(CONN_E_S,
                                                        self, 0, vacantRow))
                else:
                    self.__allocateAndSet(
                        vacantRow, 0,
                        ConnectorCell([(ConnectorCell.NORTH,
                                        ConnectorCell.SOUTH),
                                       (ConnectorCell.EAST,
                                        ConnectorCell.CENTER)],
                                      self, 0, vacantRow))
            else:
                self.__allocateAndSet(vacantRow, 0,
                                      ConnectorCell(CONN_N_S,
                                                    self, 0, vacantRow))
                if not self.__isTerminalCell(vacantRow, 1) and \
                   not self.__isVacantCell(vacantRow, 1):
                    vacantRow += 1
                    bottomConnector = ConnectorCell(CONN_N_W, self,
                                                    1, vacantRow)
                    bottomConnector.subKind = ConnectorCell.BOTTOM_IF
                    self.__allocateAndSet(vacantRow, 1, bottomConnector)
                    self.__allocateAndSet(
                        vacantRow, 0,
                        ConnectorCell([(ConnectorCell.NORTH,
                                        ConnectorCell.SOUTH),
                                       (ConnectorCell.EAST,
                                        ConnectorCell.CENTER)],
                                      self, 0, vacantRow))
        else:
            if nBranch.kind == CellElement.VCANVAS:
                if yBelow:
                    self.__allocateAndSet(vacantRow, 1, nBranch)
                else:
                    self.__allocateAndSet(vacantRow, 0, nBranch)
            else:
                # This is 'else' suite
                scopeCommentRows = 0
                if nBranch.leadingComment and not self.settings.noComment:
                    scopeCommentRows += 1
                if nBranch.sideComment and not self.settings.noComment:
                    scopeCommentRows += 1

                if yBelow:
                    branchLayout = VirtualCanvas(self.settings,
                                                 1, vacantRow,
                                                 self.__validGroups,
                                                 self.__collapsedGroups, self)
                else:
                    branchLayout = VirtualCanvas(self.settings,
                                                 0, vacantRow,
                                                 self.__validGroups,
                                                 self.__collapsedGroups, self)
                    branchLayout.isIfBelowLayout = True

                if nBranch.leadingComment and not self.settings.noComment:
                    # Draw as an independent comment: insert into the layout
                    conn = ConnectorCell(CONN_N_S, branchLayout, 0, 0)
                    cItem = self.__createIndependentComment(
                        nBranch.leadingComment, branchLayout, 1, 0)
                    branchLayout.cells.append([])
                    branchLayout.cells[0].append(conn)
                    branchLayout.cells[0].append(cItem)

                if nBranch.sideComment and not self.settings.noComment:
                    # Draw as an independent comment: insert into the layout
                    rowIndex = scopeCommentRows - 1
                    conn = ConnectorCell(CONN_N_S, branchLayout, 0, rowIndex)
                    cItem = self.__createIndependentComment(
                        nBranch.sideComment, branchLayout, 1, rowIndex)
                    cItem.sideForElse = True
                    branchLayout.cells.append([])
                    branchLayout.cells[rowIndex].append(conn)
                    branchLayout.cells[rowIndex].append(cItem)

                branchLayout.isNoScope = True
                branchLayout.layoutSuite(scopeCommentRows, nBranch.suite,
                                         CellElement.NO_SCOPE, None, 0)

                if yBelow:
                    self.__allocateAndSet(vacantRow, 1, branchLayout)
                else:
                    self.__allocateAndSet(vacantRow, 0, branchLayout)

            # Finilizing connectors
            leftTerminal = self.__isTerminalCell(vacantRow, 0) or \
                           self.__isVacantCell(vacantRow, 0)
            rightTerminal = self.__isTerminalCell(vacantRow, 1) or \
                            self.__isVacantCell(vacantRow, 1)

            if leftTerminal and rightTerminal:
                pass    # No need to do anything
            elif leftTerminal:
                vacantRow += 1
                bottomConnector = ConnectorCell(CONN_N_W, self, 1, vacantRow)
                bottomConnector.subKind = ConnectorCell.BOTTOM_IF
                self.__allocateAndSet(vacantRow, 1, bottomConnector)
                self.__allocateAndSet(vacantRow, 0,
                                      ConnectorCell(CONN_E_S,
                                                    self, 0, vacantRow))
            elif rightTerminal:
                pass    # No need to do anything
            else:
                # Both are non terminal
                vacantRow += 1
                bottomConnector = ConnectorCell(CONN_N_W, self, 1, vacantRow)
                bottomConnector.subKind = ConnectorCell.BOTTOM_IF
                self.__allocateAndSet(vacantRow, 1, bottomConnector)
                self.__allocateAndSet(
                    vacantRow, 0,
                    ConnectorCell([(ConnectorCell.NORTH,
                                    ConnectorCell.SOUTH),
                                   (ConnectorCell.EAST,
                                    ConnectorCell.CENTER)],
                                  self, 0, vacantRow))

        self.dependentRegions.append((dependentRegionBegin, vacantRow))

    def setLeaderRef(self, ref):
        """Sets the leader ref for ELSE, EXCEPT and FINALLY scopes"""
        if self.cells[0][0].kind in [CellElement.ELSE_SCOPE,
                                     CellElement.EXCEPT_SCOPE,
                                     CellElement.FINALLY_SCOPE]:
            if self.cells[0][0].subKind == ScopeCellElement.TOP_LEFT:
                self.cells[0][0].leaderRef = ref
                return
        raise Exception("Logic error: cannot set the leader reference")

    def layoutModule(self, cflow):
        """Lays out a module"""
        self.isNoScope = True
        vacantRow = 0

        # Avoid glueing to the top view edge
        self.__allocateAndSet(vacantRow, 1,
                              VSpacerCell(None, self, 1, vacantRow))
        vacantRow += 1

        vacantRow = self.__allocateLeadingComment(cflow, vacantRow, 1)
        self.__allocateScope(cflow, CellElement.FILE_SCOPE, vacantRow, 0)

        # Second stage: shifts to accomadate open groups
        self.openGroupsAdjustments()

    def getInsertIndex(self, row):
        """Provides an insert index for a spacing and a row kind.

        index: 0, 1 or -1
               -1 => no need to insert
        row kind: 0, 1, 2, -1
                  0 => regular line
                  1 => group begin line
                  -1 => group end line
                  2 => nested virtual canvas which needs to be considered
        """
        for cellIndex, cell in enumerate(row):
            if cell.kind == CellElement.VCANVAS:
                if cell.isIfBelowLayout:
                    return cellIndex, 2
                continue
            if cell.scopedItem():
                if cell.subKind in [ScopeCellElement.TOP_LEFT,
                                    ScopeCellElement.DECLARATION,
                                    ScopeCellElement.DOCSTRING] or \
                        cell.kind == CellElement.SCOPE_CORNER_EDGE:
                    return -1, None
            if cell.kind == CellElement.OPENED_GROUP_BEGIN:
                return cellIndex, 1
            if cell.kind == CellElement.OPENED_GROUP_END:
                return cellIndex, -1

        # Regular row
        if self.isNoScope:
            return 0, 0
        return 1, 0

    def openGroupsAdjustments(self):
        """Adjusts the layout if needed for the open groups"""
        localOpenGroupsStackLevel = 0
        maxLocalDepth = 0
        stackMaxDepth = 0
        for row in self.cells:
            for cell in row:
                if cell.kind == CellElement.VCANVAS:
                    if cell.isIfBelowLayout:
                        # Nested needs to be considered
                        nestedStackMaxDepth = cell.openGroupsAdjustments()
                        stackMaxDepth = max(stackMaxDepth,
                                            localOpenGroupsStackLevel +
                                            nestedStackMaxDepth)
                    else:
                        # Do the shifting independent
                        cell.openGroupsAdjustments()
                elif cell.kind == CellElement.OPENED_GROUP_BEGIN:
                    localOpenGroupsStackLevel += 1
                    stackMaxDepth = max(stackMaxDepth,
                                        localOpenGroupsStackLevel)
                    maxLocalDepth = max(maxLocalDepth,
                                        localOpenGroupsStackLevel)
                elif cell.kind == CellElement.OPENED_GROUP_END:
                    localOpenGroupsStackLevel -= 1

        # Memorize the max local (not considering the nested canvaces) depth
        self.maxLocalOpenGroupDepth = maxLocalDepth

        if self.isIfBelowLayout:
            return stackMaxDepth

        self.insertOpenGroupShift(stackMaxDepth, 0)

    def insertOpenGroupShift(self, depth, insertedByUpper):
        """Inserts the horizontal open group spacers"""
        if depth <= 0:
            return

        self.maxGlobalOpenGroupDepth = depth

        # Insert the shift as required
        localOpenGroupsStackLevel = 0
        insertCountStack = []

        for rowIndex, row in enumerate(self.cells):
            insertIndex, rowKind = self.getInsertIndex(row)
            if insertIndex < 0:
                continue

            if rowKind == 1:
                # group begin row
                groupBeginCell = self.cells[rowIndex][insertIndex]
                groupBeginCell.selfAndDeeperNestLevel, groupBeginCell.selfMaxNestLevel = self.updateGroupNestLevel(
                    rowIndex + 1, row[insertIndex].groupEndRow - 1)
                groupBeginCell.selfAndDeeperNestLevel += 1
                groupBeginCell.selfMaxNestLevel += 1

                localOpenGroupsStackLevel += 1

                insertCount = depth - groupBeginCell.selfAndDeeperNestLevel - insertedByUpper
                insertCountStack.append(insertCount)
                if insertCount > 0:
                    spacer = HGroupSpacerCell(None, self, insertIndex, rowIndex)
                    spacer.count = insertCount
                    row.insert(insertIndex, spacer)

                    insertIndex += 1
                    self.__adjustRowAddresses(row, insertIndex)
            elif rowKind == -1:
                # group end row
                groupEndCell = self.cells[rowIndex][insertIndex]
                insertCount = insertCountStack.pop()

                if insertCount > 0:
                    spacer = HGroupSpacerCell(None, self, insertIndex, rowIndex)
                    spacer.count = insertCount
                    # The group begin column has been shifted 1 cell so the
                    # reference address needs to be adjusted.
                    groupEndCell.groupBeginColumn += 1
                    row.insert(insertIndex, spacer)

                    insertIndex += 1
                    self.__adjustRowAddresses(row, insertIndex)
                localOpenGroupsStackLevel -= 1

                groupBeginCell = self.cells[groupEndCell.groupBeginRow][groupEndCell.groupBeginColumn]
                groupEndCell.selfAndDeeperNestLevel = groupBeginCell.selfAndDeeperNestLevel
            elif rowKind == 2:
                # there is a nested canvas to consider
                insertCount = 0
                if row[insertIndex].isOuterIfLayout:
                    insertCount = max(self.maxLocalOpenGroupDepth - 1,
                                      localOpenGroupsStackLevel, 0)

                    if insertCount > 0:
                        spacer = HGroupSpacerCell(None, self, insertIndex,
                                                  rowIndex)
                        spacer.count = insertCount
                        row.insert(insertIndex, spacer)

                        insertIndex += 1
                        self.__adjustRowAddresses(row, insertIndex)

                        # To have the canvas size calculated properly
                        spacerAfter = HGroupSpacerCell(None, self,
                                                       len(row), rowIndex)
                        spacerAfter.count = spacer.count
                        row.append(spacerAfter)

                row[insertIndex].insertOpenGroupShift(
                    depth, insertedByUpper + insertCount)
            else:
                # regular row
                if self.isOuterIfLayout:
                    # It could be a connector or an if block. A shift is used
                    # instead of inserting a spacer to avoid problems with
                    # dependent region rendering
                    for index in range(insertIndex, len(row)):
                        row[index].hShift = depth - insertedByUpper
                else:
                    spacer = HGroupSpacerCell(None, self,
                                              insertIndex, rowIndex)
                    spacer.count = depth - insertedByUpper
                    row.insert(insertIndex, spacer)

                    insertIndex += 1
                    self.__adjustRowAddresses(row, insertIndex)

                    # To have the canvas size calculated properly
                    spacerAfter = HGroupSpacerCell(None, self,
                                                   len(row), rowIndex)
                    spacerAfter.count = spacer.count
                    row.append(spacerAfter)

    def updateGroupNestLevel(self, startRow, endRow):
        """startRow and endRow are inclusive"""
        maxCurrentAndDeeper = 0
        maxCurrentLevel = 0
        currentLevel = 0
        for rowIndex in range(startRow, endRow + 1):
            for cell in self.cells[rowIndex]:
                if cell.kind == CellElement.OPENED_GROUP_BEGIN:
                    currentLevel += 1
                    maxCurrentAndDeeper = max(maxCurrentAndDeeper,
                                              currentLevel)
                    maxCurrentLevel = max(maxCurrentLevel, currentLevel)
                    cell.selfAndDeeperNestLevel, cell.selfMaxNestLevel = self.updateGroupNestLevel(
                        rowIndex + 1, cell.groupEndRow - 1)
                    cell.selfAndDeeperNestLevel += 1
                    cell.selfMaxNestLevel += 1
                elif cell.kind == CellElement.OPENED_GROUP_END:
                    currentLevel -= 1
                elif cell.kind == CellElement.VCANVAS:
                    if cell.isIfBelowLayout:
                        level, _ = cell.updateGroupNestLevel(
                            0, len(cell.cells) - 1)
                        maxCurrentAndDeeper = max(maxCurrentAndDeeper,
                                                  currentLevel + level)
        return maxCurrentAndDeeper, maxCurrentLevel

    @staticmethod
    def __adjustRowAddresses(row, index):
        lastIndex = len(row) - 1
        while index <= lastIndex:
            row[index].addr[0] += 1
            index += 1

    def layout(self, cflow, scopeKind, rowsToAllocate=1):
        """Does the layout"""
        self.__currentCF = cflow
        self.__currentScopeClass = _scopeToClass[scopeKind]

        vacantRow = 0

        # Allocate the scope header
        self.__allocateCell(vacantRow, 0, False)
        self.cells[vacantRow][0] = self.__currentScopeClass(
            cflow, self, 0, vacantRow, ScopeCellElement.TOP_LEFT)
        self.linesInHeader += 1
        vacantRow += 1
        self.__allocateCell(vacantRow, 1)
        self.cells[vacantRow][1] = self.__currentScopeClass(
            cflow, self, 1, vacantRow, ScopeCellElement.DECLARATION)
        self.linesInHeader += 1

        if hasattr(cflow, "sideComment"):
            if cflow.sideComment and not self.settings.noComment:
                self.__allocateCell(vacantRow - 1, 2)
                self.cells[vacantRow - 1][2] = ScopeVSideEdge(cflow, self, 2,
                                                              vacantRow - 1)
                self.__allocateCell(vacantRow, 2)
                self.cells[vacantRow][2] = self.__currentScopeClass(
                    cflow, self, 2, vacantRow, ScopeCellElement.COMMENT)

        vacantRow += 1
        if hasattr(cflow, "docstring"):
            if cflow.docstring and not self.settings.noDocstring:
                if cflow.docstring.getDisplayValue():
                    self.__allocateCell(vacantRow, 1)
                    self.cells[vacantRow][1] = self.__currentScopeClass(
                        cflow, self, 1, vacantRow, ScopeCellElement.DOCSTRING)
                    vacantRow += 1
                    self.linesInHeader += 1

        # Spaces after the header to avoid glueing the flow chart to the header
        self.__allocateAndSet(vacantRow, 1,
                              VSpacerCell(None, self, 1, vacantRow))
        vacantRow += 1

        # Handle the content of the scope
        if scopeKind == CellElement.DECOR_SCOPE:
            # no suite, just reserve the required rows
            while rowsToAllocate > 0:
                self.__allocateCell(vacantRow, 0)
                vacantRow += 1
                rowsToAllocate -= 1
        else:
            # walk the suite
            # no changes in the scope kind or control flow object
            leadingCMLComments = None
            if scopeKind == CellElement.FILE_SCOPE:
                leadingCMLComments = cflow.leadingCMLComments
            vacantRow = self.layoutSuite(vacantRow, cflow.suite, None, None, 1,
                                         leadingCMLComments)

        # Allocate the scope footer
        self.__allocateCell(vacantRow, 0, False)
        self.cells[vacantRow][0] = ScopeSpacer(cflow, self, 0, vacantRow)

        self.__currentCF = None

    def __dependentRegion(self, rowIndex):
        """True if it is a dependent region"""
        if self.dependentRegions:
            return self.dependentRegions[0][0] == rowIndex
        return False

    def __getRangeMaxColumns(self, start, end):
        """Provides the max columns"""
        maxColumns = 0
        while start <= end:
            maxColumns = max(maxColumns, len(self.cells[start]))
            start += 1
        return maxColumns

    def __renderRegion(self, openGroups):
        """Renders a region where the rows affect each other"""
        start, end = self.dependentRegions.pop(0)
        maxColumns = self.__getRangeMaxColumns(start, end)

        # Detect the region tail comment cells
        index = start
        while index <= end:
            row = self.cells[index]
            if row:
                if row[-1].kind in [CellElement.LEADING_COMMENT,
                                    CellElement.INDEPENDENT_COMMENT,
                                    CellElement.SIDE_COMMENT,
                                    CellElement.ABOVE_COMMENT,
                                    CellElement.LEADING_MINIMIZED_COMMENT,
                                    CellElement.INDEPENDENT_MINIMIZED_COMMENT,
                                    CellElement.SIDE_MINIMIZED_COMMENT,
                                    CellElement.ABOVE_MINIMIZED_COMMENT]:
                    row[-1].tailComment = True
            index += 1

        for column in range(maxColumns):
            maxColumnWidth = 0
            index = start
            while index <= end:
                row = self.cells[index]
                if column < len(row):
                    row[column].render()
                    if not row[column].scopedItem():
                        if row[column].isComment() or row[column].isCMLDoc():
                            row[column].adjustWidth()
                    if not row[column].tailComment:
                        maxColumnWidth = max(row[column].width, maxColumnWidth)
                index += 1

            index = start
            while index <= end:
                row = self.cells[index]
                if column < len(row):
                    if not row[column].tailComment:
                        row[column].width = maxColumnWidth
                index += 1

        # Update the row height and calculate the row width.
        # It also catches the open groups.
        index = start
        while index <= end:
            maxHeight = 0
            row = self.cells[index]
            rowWidth = 0
            for cell in row:
                if cell.kind == CellElement.OPENED_GROUP_END:
                    openGroups.append([cell.groupBeginRow,
                                       cell.groupBeginColumn,
                                       index])
                maxHeight = max(cell.height, maxHeight)
                rowWidth += cell.width
            self.width = max(self.width, rowWidth)
            for cell in row:
                cell.height = maxHeight
                if cell.kind == CellElement.VCANVAS:
                    if not cell.hasScope():
                        cell.adjustLastCellHeight(maxHeight)
            self.height += maxHeight
            index += 1

        return end

    def hasScope(self):
        """True if it has a scope"""
        try:
            return self.cells[0][0].scopedItem()
        except:
            return False

    def render(self):
        """Preforms rendering for all the cells"""
        self.width = 0
        self.height = 0

        openGroups = []
        maxRowIndex = len(self.cells) - 1
        index = 0
        while index <= maxRowIndex:
            if self.__dependentRegion(index):
                index = self.__renderRegion(openGroups)
            else:
                row = self.cells[index]
                maxHeight = 0
                for cell in row:
                    if cell.kind == CellElement.OPENED_GROUP_END:
                        openGroups.append([cell.groupBeginRow,
                                           cell.groupBeginColumn,
                                           index])

                    _, height = cell.render()
                    maxHeight = max(maxHeight, height)
                    if not cell.scopedItem():
                        if cell.isComment() or cell.isCMLDoc():
                            cell.adjustWidth()
                for cell in row:
                    cell.height = maxHeight
                    if cell.kind == CellElement.VCANVAS:
                        if not cell.hasScope():
                            cell.adjustLastCellHeight(maxHeight)
                self.height += maxHeight
            index += 1

        # Second pass for the groups
        for groupBeginRow, groupBeginColumn, groupEndRow in openGroups:
            group = self.cells[groupBeginRow][groupBeginColumn]
            group.groupHeight = 0
            group.groupWidth = 0
            for row in range(groupBeginRow + 1, groupEndRow):
                group.groupHeight += self.cells[row][0].height
                rowWidth = 0
                for column in range(groupBeginColumn, len(self.cells[row])):
                    cell = self.cells[row][column]
                    if cell.kind != CellElement.H_GROUP_SPACER:
                        rowWidth += cell.width
                group.groupWidth = max(group.groupWidth, rowWidth)
            group.groupWidth += (group.selfMaxNestLevel - 1) * 4 * self.settings.openGroupHSpacer

        # There is no need to add spacing at the right hand side for groups.
        # The appropriate number of spacers were added for canvases and regular
        # rows so no adjustments here.

        # Third pass for if branch adjustments
        for rowIndex, row in enumerate(self.cells):
            for cellIndex, cell in enumerate(row):
                if cell.kind == CellElement.IF:
                    cellBelow = self.cells[rowIndex + 1][cellIndex]
                    spare = cellBelow.width - cellBelow.minWidth
                    minShift = 2 * self.settings.hCellPadding + self.settings.mainLine + self.settings.ifWidth
                    if spare > minShift:
                        cellBelowRight = self.cells[rowIndex + 1][cellIndex + 1]
                        if cellBelowRight.kind == CellElement.VCANVAS:
                            # The rhs branch will be shifted to the left
                            if spare < cellBelowRight.width:
                                # The RHS branch is wider than the available shift
                                # width so use it all
                                shift = spare
                            else:
                                # The RHS branch is narrower than the available
                                # shift width so use only the cell width
                                shift = cellBelowRight.width
                            # Signal the IF cell that there will be no RHS
                            # connector but a bottom instead
                            cell.needHConnector = False
                            cell.rhsShift = shift

                            cellBelow.width -= shift
                            # Delete the RHS if connector cell
                            del self.cells[rowIndex][cellIndex + 1]

                            try:
                                # The if side comment heeds to be re-rendered
                                # because it is not region dependent anymore
                                sideCommentCell = self.cells[rowIndex][cellIndex + 1]
                                if sideCommentCell.kind in [CellElement.SIDE_COMMENT,
                                                            CellElement.SIDE_MINIMIZED_COMMENT]:
                                    oldMinHeight = sideCommentCell.minHeight
                                    oldHeight = sideCommentCell.height
                                    sideCommentCell.render()
                                    sideCommentCell.minHeight = oldMinHeight
                                    sideCommentCell.height = oldHeight
                                    # The rhs connector has been deleted, so the
                                    # x address is changed
                                    sideCommentCell.addr = [cellIndex + 1, rowIndex]
                            except:
                                pass

                            # Finishing connector if so, should be adjusted too
                            try:
                                closingCell = self.cells[rowIndex + 2][cellIndex]
                                if closingCell.kind == CellElement.CONNECTOR:
                                    if closingCell.hasVertical() and closingCell.hasHorizontal():
                                        closingCell.width -= shift
                            except:
                                pass
                            break

        # The if cells adjustments may have deleted connectors and adjusted
        # cell width, so recalculate the canves width
        self.width = 0
        for row in self.cells:
            totalWidth = 0
            for cell in row:
                totalWidth += cell.width
            self.width = max(self.width, totalWidth)

        if self.hasScope():
            # Right hand side vertical part
            self.width += self.settings.scopeRectRadius + self.settings.hCellPadding

        self.minWidth = self.width
        self.minHeight = self.height
        return (self.width, self.height)

    def adjustLastCellHeight(self, maxHeight):
        """Adjusts the cell height if needed.

        The last cell in the first column of the non-scope virtual canvas
        may need to be adjusted to occupy the whole row hight in the upper
        level canvas. This happens mostly in 'if' statements.
        """
        if not self.cells:
            return

        # The last items could be group ends so we need to find the last
        # meaningful row.
        rowIndex = len(self.cells) - 1
        while self.cells[rowIndex][-1].kind == CellElement.OPENED_GROUP_END:
            rowIndex -= 1

        allExceptLastMeaningfulHeight = 0
        for index in range(len(self.cells)):
            if index != rowIndex:
                allExceptLastMeaningfulHeight += self.cells[index][0].height

        # Update the height for all the cells in the last row
        for cell in self.cells[rowIndex]:
            if allExceptLastMeaningfulHeight + cell.height < maxHeight:
                cell.height = maxHeight - allExceptLastMeaningfulHeight
                if cell.kind == CellElement.VCANVAS:
                    if not cell.hasScope():
                        cell.adjustLastCellHeight(cell.height)

    def setEditor(self, editor):
        """Provides the editor counterpart"""
        self.editor = editor

    def draw(self, scene, baseX, baseY):
        """Draws the diagram on the real canvas"""
        self.baseX = baseX
        self.baseY = baseY
        currentY = baseY
        for row in self.cells:
            if not row:
                continue
            height = row[0].height
            currentX = baseX
            for cell in row:
                if self.settings.debug:
                    if cell.kind == CellElement.VCANVAS:
                        rect = Rectangle(self, currentX + 1, currentY + 1,
                                         cell.width - 2, cell.height -2)
                        rect.pen = QPen(QColor(255, 0, 0, 255))
                        rect.brush = QBrush(QColor(255, 0, 0, 127))
                        rect.setToolTip('Canvas ' + str(cell.width) + 'x' +
                                        str(cell.height))
                        scene.addItem(rect)
                    else:
                        rect = Rectangle(self, currentX, currentY,
                                         cell.width, cell.height)
                        rect.pen = QPen(QColor(0, 255, 0, 255))
                        rect.brush = QBrush(QColor(0, 255, 0, 127))
                        rect.setToolTip('Item ' + str(cell) +
                                        ' ' + str(cell.kind))
                        scene.addItem(rect)
                cell.draw(scene, currentX, currentY)
                currentX += cell.width
            currentY += height


def formatFlow(s):
    """Reformats the control flow output"""
    result = ""
    shifts = []     # positions of opening '<'
    pos = 0         # symbol position in a line
    nextIsList = False

    def IsNextList(index, maxIndex, buf):
        if index == maxIndex:
            return False
        if buf[index + 1] == '<':
            return True
        if index < maxIndex - 1:
            if buf[index + 1] == '\n' and buf[index + 2] == '<':
                return True
        return False

    maxIndex = len(s) - 1
    for index in range(len(s)):
        sym = s[index]
        if sym == "\n":
            lastShift = shifts[-1]
            result += sym + lastShift * " "
            pos = lastShift
            if index < maxIndex:
                if s[index + 1] not in "<>":
                    result += " "
                    pos += 1
            continue
        if sym == "<":
            if nextIsList == False:
                shifts.append(pos)
            else:
                nextIsList = False
            pos += 1
            result += sym
            continue
        if sym == ">":
            shift = shifts[-1]
            result += '\n'
            result += shift * " "
            pos = shift
            result += sym
            pos += 1
            if IsNextList(index, maxIndex, s):
                nextIsList = True
            else:
                del shifts[-1]
                nextIsList = False
            continue
        result += sym
        pos += 1
    return result



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Provides the storage for the last position in a file"""

import os.path
from .fileutils import loadJSON, saveJSON


class FilePositions:

    """Loads/stores/saves the last position in files"""

    def __init__(self):
        # file name -> (line, pos, first visible, horizontalPos, verticalPos)
        self.__filePos = {}
        self.__fpFileName = None

    def reset(self):
        """Un-binds from the file system"""
        self.__filePos = {}
        self.__fpFileName = None

    def setup(self, dirName):
        """Binds the parameters to a disk file"""
        # Just in case - flush the previous data if they were bound
        FilePositions.save(self)

        dirName = os.path.realpath(dirName)
        if not dirName.endswith(os.path.sep):
            dirName += os.path.sep
        if not os.path.isdir(dirName):
            raise Exception('Directory name is expected for files '
                            'positions. The given ' + dirName + ' is not.')

        self.__fpFileName = dirName + 'lastpositions.json'
        if os.path.exists(self.__fpFileName):
            FilePositions.load(self)

    def load(self):
        """Loads the saved positions file"""
        if self.__fpFileName:
            self.__filePos = loadJSON(self.__fpFileName,
                                      'file editing positions', {})

    def save(self):
        """Saves the positions into a file"""
        if self.__fpFileName:
            saveJSON(self.__fpFileName, self.__filePos,
                     'file editing positions')

    def getFilePosition(self, fileName):
        """Provides the position or (-1,-1,-1,-1,-1) if not found"""
        return self.__filePos.get(fileName, (-1, -1, -1, -1, -1))

    def updateFilePosition(self, fileName, line, pos, firstLine,
                           horizontalPos, verticalPos):
        """Updates the position for the file"""
        self.__filePos[fileName] = (line, pos, firstLine,
                                    horizontalPos, verticalPos)
        FilePositions.save(self)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 6 and adopted for codimension.
# Original copyright:
# Copyright (c) 2002 - 2017 Detlev Offenbach <detlev@die-offenbachs.de>
#


"""
Module implementing the multithreaded version of the debug client
"""

import sys
import traceback
import time
from base_cdm_dbg import DebugBase
from clientbase_cdm_dbg import DebugClientBase
from threadextension_cdm_dbg import ThreadExtension


CLIENT_DEBUG = True


class DebugClient(DebugClientBase, DebugBase, ThreadExtension):

    """
    Class implementing the client side of the debugger
    """

    def __init__(self):
        DebugClientBase.__init__(self)
        DebugBase.__init__(self, self)
        ThreadExtension.__init__(self)


# We are normally called by the debugger to execute directly.

if __name__ == '__main__':
    debugClient = DebugClient()
    try:
        debugClient.main()
    except Exception as exc:
        if CLIENT_DEBUG:
            print(traceback.format_exc(), file=sys.__stderr__)
            if sys.__stderr__ != sys.stderr:
                print(traceback.format_exc(), file=sys.stderr)
            # The delay is for a socket data
            time.sleep(4)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4 and adopted for codimension.
# Original copyright:
# Copyright (c) 2002 - 2012 Detlev Offenbach <detlev@die-offenbachs.de>
#


"""
Module implementing the debug base class
"""

import sys
import os
import types
import atexit
import inspect
import ctypes
import time
from inspect import CO_GENERATOR
from bp_wp_cdm_dbg import Breakpoint, Watch
from cdm_dbg_utils import formatArgValues, getArgValues, printerr
import _thread


RECURSION_LIMIT = 64


def setRecursionLimit(limit):
    " Sets the recursion limit "
    global RECURSION_LIMIT
    RECURSION_LIMIT = limit


class DebugBase(object):

    """
    Base class of the debugger - the part which is external to IDE.
    """

    # Don't thrust distutils.sysconfig.get_python_lib: possible case mismatch
    # on Windows
    lib = os.path.dirname(inspect.__file__)

    # Tuple required because it's accessed a lot of times by startswith method
    pathsToSkip = ('<', os.path.dirname(__file__), inspect.__file__[:-1])
    filesToSkip = {}

    # Cache for fixed file names
    _fnCache = {}

    # Stop all timers, when greenlets are used
    pollTimerEnabled = True

    def __init__(self, dbgClient):
        self._dbgClient = dbgClient

        # Some informations about the thread
        self.isMainThread = False
        self.quitting = False
        self.id = -1
        self.name = ''

        self.tracePythonLibs(0)

        # Special handling of a recursion error
        self.skipFrames = 0

        self.isBroken = False
        self.cFrame = None

        # current frame we are at
        self.currentFrame = None

        # frames, where we want to stop or release debugger
        self.stopframe = None
        self.returnframe = None
        self.stop_everywhere = False

        self.__recursionDepth = -1
        self.setRecursionDepth(inspect.currentframe())

        # background task to periodicaly check for client interactions
        self.eventPollFlag = False
        self.timer = _thread.start_new_thread(self.__eventPollTimer, ())

        # provide a hook to perform a hard breakpoint
        # Use it like this:
        # if hasattr(sys, 'breakpoint): sys.breakpoint()
        sys.breakpoint = self.set_trace
        if sys.version_info[:2] >= (3, 7):
            sys.breakpointhook = self.set_trace

    def __eventPollTimer(self):
        """Sets a flag every 0.5 sec to check for new messages"""
        while DebugBase.pollTimerEnabled:
            time.sleep(0.5)
            self.eventPollFlag = True

        self.eventPollFlag = False

    def getCurrentFrame(self):
        """Provides the current frame"""
        if self.quitting:
            return None
        return self.currentFrame

    def getFrameLocals(self, frmnr=0):
        """Provides the locals dictionary of the current frame"""
        f = self.currentFrame
        while f is not None and frmnr > 0:
            f = f.f_back
            frmnr -= 1
        return f.f_locals

    def storeFrameLocals(self, frmnr=0):
        """Stores the locals into the frame.

        Thus an access to frame.f_locals returns the last data
        """
        cf = self.currentFrame
        while cf is not None and frmnr > 0:
            cf = cf.f_back
            frmnr -= 1

        try:
            if '__pypy__' in sys.builtin_module_names:
                import __pypy__
                __pypy__.locals_to_fast(cf)
                return
        except Exception:
            pass

        ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(cf),
                                              ctypes.c_int(0))

    def step(self, traceMode):
        """Performs step in this thread"""
        if traceMode:
            self.set_step()
        else:
            self.set_next(self.currentFrame)

    def stepOut(self):
        """Performs a step out of the current call"""
        self.set_return(self.currentFrame)

    def go(self, special):
        """Resumes the thread"""
        self.set_continue(special)

    def setRecursionDepth(self, frame):
        """Determines the current recursion depth"""
        self.__recursionDepth = 0
        while frame is not None:
            self.__recursionDepth += 1
            frame = frame.f_back

    def profileWithRecursion(self, frame, event, arg):
        """Traces some stuff independent of the debugger trace function"""
        if event == 'return':
            self.cFrame = frame.f_back
            self.__recursionDepth -= 1
            if self._dbgClient.callTraceEnabled:
                self.__sendCallTrace(event, frame, self.cFrame)
        elif event == 'call':
            if self._dbgClient.callTraceEnabled:
                self.__sendCallTrace(event, self.cFrame, frame)
            self.cFrame = frame
            self.__recursionDepth += 1
            if self.__recursionDepth > RECURSION_LIMIT:
                raise RuntimeError('maximum recursion depth exceeded\n'
                                   '(offending frame is two down the stack)')

    def profile(self, frame, event, arg):
        """Traces some stuff independant of the debugger trace function"""
        if event == 'return':
            self.__sendCallTrace(event, frame, frame.f_back)
        elif event == 'call':
            self.__sendCallTrace(event, frame.f_back, frame)

    def __sendCallTrace(self, event, fromFrame, toFrame):
        """Sends a call/return trace"""
        if not self.__skipFrame(fromFrame) and not self.__skipFrame(toFrame):
            fromInfo = {
                "filename": self._dbgClient.absPath(
                    self.fix_frame_filename(fromFrame)),
                "linenumber": fromFrame.f_lineno,
                "codename": fromFrame.f_code.co_name}
            toInfo = {
                "filename": self._dbgClient.absPath(
                    self.fix_frame_filename(toFrame)),
                "linenumber": toFrame.f_lineno,
                "codename": toFrame.f_code.co_name}
            self._dbgClient.sendCallTrace(event, fromInfo, toInfo)

    def trace_dispatch(self, frame, event, arg):
        """Reimplemented from bdb.py to do some special things"""
        # give the client a chance to push through new break points.
        if self.eventPollFlag:
            self._dbgClient.eventPoll()
            self.eventPollFlag = False

            if self.quitting:
                raise SystemExit

        if event == 'line':
            if self.stop_here(frame) or self.break_here(frame):
                if (self.stop_everywhere and frame.f_back and
                        frame.f_back.f_code.co_name == "sendJSONCommand"):
                    # Just stepped into print statement, so skip these frames
                    self._set_stopinfo(None, frame.f_back)
                else:
                    self.user_line(frame)
            return self.trace_dispatch

        if event == 'call':
            if (self.stop_here(frame) or
                    self.__checkBreakInFrame(frame) or
                    Watch.WATCHES != []):
                return self.trace_dispatch
            # No need to trace this function
            return

        if event == 'return':
            if frame == self.returnframe:
                # Only true if we didn't stopped in this frame, because it's
                # belonging to the eric debugger.
                self._set_stopinfo(None, frame.f_back)
            return

        if event == 'exception':
            if not self.__skipFrame(frame):
                # When stepping with next/until/return in a generator frame,
                # skip the internal StopIteration exception (with no traceback)
                # triggered by a subiterator run with the 'yield from'
                # statement.
                if not (frame.f_code.co_flags & CO_GENERATOR and
                        arg[0] is StopIteration and arg[2] is None):
                    self.user_exception(arg)
            # Stop at the StopIteration or GeneratorExit exception when the
            # user has set stopframe in a generator by issuing a return
            # command, or a next/until command at the last statement in the
            # generator before the exception.
            elif (self.stopframe and frame is not self.stopframe and
                  self.stopframe.f_code.co_flags & CO_GENERATOR and
                  arg[0] in (StopIteration, GeneratorExit)):
                self.user_exception(arg)
            return

        if event == 'c_call':
            return
        if event == 'c_exception':
            return
        if event == 'c_return':
            return

        print('DebugBase.trace_dispatch:'
              ' unknown debugging event: ',
              repr(event))
        return self.trace_dispatch

    def set_trace(self, frame=None):
        """Starts debugging from 'frame'"""
        if frame is None:
            frame = sys._getframe().f_back  # Skip set_trace method

        if sys.version_info[0] == 2:
            stopOnHandleLine = self._dbgClient.handleLine.func_code
        else:
            stopOnHandleLine = self._dbgClient.handleLine.__code__

        frame.f_trace = self.trace_dispatch
        while frame.f_back is not None:
            # stop at erics debugger frame or a threading bootstrap
            if frame.f_back.f_code == stopOnHandleLine:
                frame.f_trace = self.trace_dispatch
                break

            frame = frame.f_back

        self.stop_everywhere = True
        sys.settrace(self.trace_dispatch)
        sys.setprofile(self._dbgClient.callTraceEnabled)

    def bootstrap(self, target, args, kwargs):
        """Bootstraps a thread"""
        try:
            # Because in the initial run method the "base debug" function is
            # set up, it's also valid for the threads afterwards.
            sys.settrace(self.trace_dispatch)

            target(*args, **kwargs)
        except Exception:
            excinfo = sys.exc_info()
            self.user_exception(excinfo, True)
        finally:
            sys.settrace(None)
            sys.setprofile(None)

    def run(self, cmd, globalsDict=None, localsDict=None, debug=True):
        """Starts a given command under debugger control"""
        if globalsDict is None:
            import __main__
            globalsDict = __main__.__dict__

        if localsDict is None:
            localsDict = globalsDict

        if not isinstance(cmd, types.CodeType):
            cmd = compile(cmd, "<string>", "exec")

        if debug:
            # First time the trace_dispatch function is called, a "base debug"
            # function has to be returned, which is called at every user code
            # function call. This is ensured by setting stop_everywhere.
            self.stop_everywhere = True
            sys.settrace(self.trace_dispatch)

        try:
            exec(cmd, globalsDict, localsDict)
            atexit._run_exitfuncs()
            self._dbgClient.progTerminated(0)
        except SystemExit:
            atexit._run_exitfuncs()
            excinfo = sys.exc_info()
            exitcode, message = self.__extractSystemExitMessage(excinfo)
            self._dbgClient.progTerminated(exitcode, message)
        except Exception:
            excinfo = sys.exc_info()
            self.user_exception(excinfo, True)
        finally:
            self.quitting = True
            sys.settrace(None)

    def _set_stopinfo(self, stopframe, returnframe):
        """Updates the frame pointers"""
        self.stopframe = stopframe
        self.returnframe = returnframe
        if returnframe is not None:
            # Ensure to be able to stop on the return frame
            returnframe.f_trace = self.trace_dispatch
        self.stop_everywhere = False

    def set_continue(self, special):
        """Stops only on next breakpoint"""
        # Here we only set a new stop frame if it is a normal continue.
        if not special:
            self._set_stopinfo(None, None)

        # Disable tracing if not started in debug mode
        if not self._dbgClient.debugging:
            sys.settrace(None)
            sys.setprofile(None)

    def set_step(self):
        """Stops after one line of code"""
        self._set_stopinfo(None, None)
        self.stop_everywhere = True

    def set_next(self, frame):
        """Stops on the next line in or below the given frame"""
        self._set_stopinfo(frame, frame.f_back)
        frame.f_trace = self.trace_dispatch

    def set_return(self, frame):
        """Stops when returning from the given frame"""
        self._set_stopinfo(None, frame.f_back)

    def move_instruction_pointer(self, lineno):
        """Moves the instruction pointer to another line"""
        try:
            self.currentFrame.f_lineno = lineno
            stack = self.getStack(self.currentFrame)
            self._dbgClient.sendResponseLine(stack)
        except Exception as exc:
            printerr(str(exc))

    def set_quit(self):
        """Quits"""
        sys.setprofile(None)
        self.stopframe = None
        self.returnframe = None
        for debugThread in self._dbgClient.threads.values():
            debugThread.quitting = True

    def fix_frame_filename(self, frame):
        """Fixups the filename for a given frame"""
        # get module name from __file__
        fn = frame.f_globals.get('__file__')
        try:
            return self._fnCache[fn]
        except KeyError:
            if fn is None:
                return frame.f_code.co_filename

            absFilename = os.path.abspath(fn)
            if absFilename.endswith(('.pyc', '.pyo', '.pyd')):
                fixedName = absFilename[:-1]
                if not os.path.exists(fixedName):
                    fixedName = absFilename
            else:
                fixedName = absFilename
            # update cache
            self._fnCache[fn] = fixedName
            return fixedName

    def __checkBreakInFrame(self, frame):
        """Checks if the function/method has a line number which is a bp"""
        try:
            return Breakpoint.BREAK_IN_FRAME_CACHE[
                frame.f_globals.get('__file__'),
                frame.f_code.co_firstlineno]
        except KeyError:
            filename = self.fix_frame_filename(frame)
            if filename not in Breakpoint.BREAK_IN_FILE:
                Breakpoint.BREAK_IN_FRAME_CACHE[
                    frame.f_globals.get('__file__'),
                    frame.f_code.co_firstlineno] = False
                return False
            lineNo = frame.f_code.co_firstlineno
            lineNumbers = [lineNo]

            if sys.version_info[0] == 2:
                co_lnotab = map(ord, frame.f_code.co_lnotab[1::2])
            else:
                co_lnotab = frame.f_code.co_lnotab[1::2]

            # No need to handle special case if a lot of lines between
            # (e.g. closure), because the additional lines won't cause a bp
            for co_lno in co_lnotab:
                lineNo += co_lno
                lineNumbers.append(lineNo)

            for bp in Breakpoint.BREAK_IN_FILE[filename]:
                if bp in lineNumbers:
                    Breakpoint.BREAK_IN_FRAME_CACHE[
                        frame.f_globals.get('__file__'),
                        frame.f_code.co_firstlineno] = True
                    return True
            Breakpoint.BREAK_IN_FRAME_CACHE[
                frame.f_globals.get('__file__'),
                frame.f_code.co_firstlineno] = False
            return False

    def break_here(self, frame):
        """Reimplemented from bdb.py to fix the filename from the frame"""
        filename = self.fix_frame_filename(frame)
        if (filename, frame.f_lineno) in Breakpoint.BREAKS:
            bp, flag = Breakpoint.effectiveBreak(
                filename, frame.f_lineno, frame)
            if bp:
                # flag says ok to delete temp. bp
                if flag and bp.temporary:
                    self.__do_clearBreak(filename, frame.f_lineno)
                return True

        if Watch.WATCHES != []:
            bp, flag = Watch.effectiveWatch(frame)
            if bp:
                # flag says ok to delete temp. watch
                if flag and bp.temporary:
                    self.__do_clearWatch(bp.cond)
                return True

        return False

    def __do_clearBreak(self, filename, lineno):
        """Clears a temporary breakpoint"""
        Breakpoint.clear_break(filename, lineno)
        self._dbgClient.sendClearTemporaryBreakpoint(filename, lineno)

    def __do_clearWatch(self, cond):
        """Clears a temporary watch expression"""
        Watch.clear_watch(cond)
        self._dbgClient.sendClearTemporaryWatch(cond)

    def getStack(self, frame=None, applyTrace=False):
        """Provides the stack"""
        if frame is None:
            fr = self.getCurrentFrame()
        elif type(frame) == list:
            fr = frame.pop(0)
        else:
            fr = frame

        stack = []
        while fr is not None:
            if applyTrace:
                # Reset the trace function so we can be sure
                # to trace all functions up the stack... This gets around
                # problems where an exception/breakpoint has occurred
                # but we had disabled tracing along the way via a None
                # return from dispatch_call
                fr.f_trace = self.trace_dispatch

            fname = self._dbgClient.absPath(self.fix_frame_filename(fr))
            # Always show at least one stack frame, even if it's from
            # codimension
            if stack and os.path.basename(fname).startswith(
                    ('base_cdm_dbg.py', "clientbase_cdm_dbg.py",
                     "threadextension_cdm_dbg.py", "threading.py")):
                break

            fline = fr.f_lineno
            ffunc = fr.f_code.co_name

            if ffunc == '?':
                ffunc = ''

            if ffunc and not ffunc.startswith("<"):
                argInfo = getArgValues(fr)
                try:
                    fargs = formatArgValues(
                        argInfo.args, argInfo.varargs,
                        argInfo.keywords, argInfo.locals)
                except Exception:
                    fargs = ""
            else:
                fargs = ""

            stack.append([fname, fline, ffunc, fargs])

            # is it a stack frame or exception list?
            if type(frame) == list:
                if frame != []:
                    fr = frame.pop(0)
                else:
                    fr = None
            else:
                fr = fr.f_back
        return stack

    def user_line(self, frame):
        """Reimplemented to the execution a particular line"""
        # We never stop on line 0.
        if frame.f_lineno == 0:
            return

        self.isBroken = True
        self.currentFrame = frame
        stack = self.getStack(frame, applyTrace=True)

        self._dbgClient.lockClient()
        self._dbgClient.currentThread = self
        self._dbgClient.currentThreadExec = self

        self._dbgClient.sendResponseLine(stack)
        self._dbgClient.eventLoop()

        self.isBroken = False
        self._dbgClient.unlockClient()

    def user_exception(self, excinfo, unhandled=False):
        """Reimplemented to report an exception to the debug server"""
        exctype, excval, exctb = excinfo

        if ((exctype in [GeneratorExit, StopIteration] and unhandled == False)
            or exctype == SystemExit):
            # ignore these
            return

        if exctype in [SyntaxError, IndentationError]:
            try:
                # tuple could only occure on Python 2, but not always!
                if type(excval) == tuple:
                    message, details = excval
                    filename, lineno, charno, text = details
                else:
                    message = excval.msg
                    filename = excval.filename
                    lineno = excval.lineno
                    charno = excval.offset

                if filename is None:
                    realSyntaxError = False
                else:
                    if charno is None:
                        charno = 0

                    filename = os.path.abspath(filename)
                    realSyntaxError = os.path.exists(filename)

            except (AttributeError, ValueError):
                message = ""
                filename = ""
                lineno = 0
                charno = 0
                realSyntaxError = True

            if realSyntaxError:
                self._dbgClient.sendSyntaxError(
                    message, filename, lineno, charno)
                self._dbgClient.eventLoop()
                return

        self.skipFrames = 0
        if (exctype == RuntimeError and
                str(excval).startswith('maximum recursion depth exceeded') or
                sys.version_info >= (3, 5) and
                exctype == RecursionError):
            excval = 'maximum recursion depth exceeded'
            depth = 0
            tb = exctb
            while tb:
                tb = tb.tb_next

                if (tb and tb.tb_frame.f_code.co_name == 'trace_dispatch' and
                        __file__.startswith(tb.tb_frame.f_code.co_filename)):
                    depth = 1
                self.skipFrames += depth

            # always 1 if running without debugger
            self.skipFrames = max(1, self.skipFrames)

        exctype = self.__extractExceptionName(exctype)

        if excval is None:
            excval = ''

        if unhandled:
            exctypetxt = "Unhandled {0!s}".format(str(exctype))
        else:
            exctypetxt = str(exctype)

        # Don't step into libraries, which are used by our debugger methods
        if exctb is not None:
            self.stop_everywhere = False

        self.isBroken = True

        stack = []
        if exctb:
            frlist = self.__extract_stack(exctb)
            frlist.reverse()

            self.currentFrame = frlist[0]
            stack = self.getStack(frlist[self.skipFrames:])

        self._dbgClient.lockClient()
        self._dbgClient.currentThread = self
        self._dbgClient.currentThreadExec = self
        self._dbgClient.sendException(exctypetxt, str(excval), stack)
        self._dbgClient.dumpThreadList()

        if exctb is not None:
            # When polling kept enabled, it isn't possible to resume after an
            # unhandled exception without further user interaction.
            self._dbgClient.eventLoop(True)

        self.skipFrames = 0

        self.isBroken = False
        stop_everywhere = self.stop_everywhere
        self.stop_everywhere = False
        self.eventPollFlag = False
        self._dbgClient.unlockClient()
        self.stop_everywhere = stop_everywhere

    @staticmethod
    def __extractExceptionName(exctype):
        """Extracts the exception name given the exception type object"""
        return str(exctype).replace("<class '", "").replace("'>", "")

    def __extract_stack(self, exctb):
        """Returns a list of stack frames"""
        tb = exctb
        stack = []
        while tb is not None:
            stack.append(tb.tb_frame)
            tb = tb.tb_next
        tb = None
        return stack

    def __extractSystemExitMessage(self, excinfo):
        """Provides the SystemExit code and message"""
        exctype, excval, exctb = excinfo
        if excval is None:
            exitcode = 0
            message = ""
        elif isinstance(excval, str):
            exitcode = 1
            message = excval
        elif isinstance(excval, bytes):
            exitcode = 1
            message = excval.decode()
        elif isinstance(excval, int):
            exitcode = excval
            message = ""
        elif isinstance(excval, SystemExit):
            code = excval.code
            if isinstance(code, str):
                exitcode = 1
                message = code
            elif isinstance(code, bytes):
                exitcode = 1
                message = code.decode()
            elif isinstance(code, int):
                exitcode = code
                message = ""
            else:
                exitcode = 1
                message = str(code)
        else:
            exitcode = 1
            message = str(excval)

        return exitcode, message

    def stop_here(self, frame):
        """Reimplemented to filter out debugger files"""
        if self.__skipFrame(frame):
            return False

        return (self.stop_everywhere or
                frame is self.stopframe or
                frame is self.returnframe)

    def tracePythonLibs(self, enable):
        """Updates the settings to trace into Python libraries"""
        pathsToSkip = list(self.pathsToSkip)
        # don't trace into Python library?
        if enable:
            pathsToSkip = [x for x in pathsToSkip if not x.endswith(
                ("site-packages", "dist-packages", self.lib))]
        else:
            pathsToSkip.append(self.lib)
            localLib = [x for x in sys.path
                        if x.endswith(("site-packages", "dist-packages")) and
                        not x.startswith(self.lib)]
            pathsToSkip.extend(localLib)

        self.pathsToSkip = tuple(set(pathsToSkip))

    def __skipFrame(self, frame):
        """Filters out debugger files"""
        try:
            return self.filesToSkip[frame.f_code.co_filename]
        except KeyError:
            ret = frame.f_code.co_filename.startswith(self.pathsToSkip)
            self.filesToSkip[frame.f_code.co_filename] = ret
            return ret
        except AttributeError:
            # if frame is None
            return True
